#!/usr/bin/env python
# weili@jimmy.harvard.edu

# Summarize the hmm data and print out a good bed file.

import sys, math, time
from numarray import *


class HMM(object):
    ''' HMM file objects, self.data = array(( num_probe, 3)), self.chr'''
    def __init__(self, filename):
        self.chr  = ''
        self.data = []

        print >> sys.stderr, "Reading HMM file ", filename
        for x in open(filename):
            if x[:4] == '#chr':
                self.chr = x[1:-1]
            if not x[0].isdigit():
                continue

            #real data
            x = x.split()
            try:                        #maybe outlier probes
                self.data.append([float(i) for i in x  ])
            except:
                continue
        self.data = array(self.data)


def main(options, argv):
    outfile = open(options.outfile , 'w')
    sys.stdout = outfile                            #replace the stdout as external file
    minLen = 35 * (options.pcutoff-1)

    print "#", time.ctime(),"Running parametres:  " , " ".join(sys.argv)
    print "#Chr\tSt\tEnd\tBlk_Size\tScore"
    id = 1      #Blk id

    for infile in argv:
        hmm =HMM(infile)
        enriched = (hmm.data[:,1] >= options.Tcutoff) * (hmm.data[:,2]<= options.Ccutoff)
        enriched = enriched.astype(Bool)
        enriched = hmm.data[enriched]

        for i, x in enumerate(enriched):
            #start and init
            if i == 0 or x[0] - enriched[i-1][0] > options.InterProbe:
                st = x[0]
                ipon = controlon = summary = count = 0

            #check the Cpeak and Tpeak cutoff
            if x[1] >= options.Tpeak:
                ipon = 1
            if x[2] <= options.Cpeak:
                controlon = 1

            summary += x[1]
            count += 1

            #end
            if i == enriched.shape[0]-1 or enriched[i+1][0] - x[0] > options.InterProbe:
                end = x[0]
                if ipon and controlon and ((end-st) > minLen):
                    print "%s\t%d\t%d\tBlk%d_%d\t%.1f" % (hmm.chr, st, end, id, end-st, summary/count)
                    id += 1

#read program options
if __name__ == '__main__':
    import optparse

    parser = optparse.OptionParser("%prog options hmmfile1 hmmfile2 hmmfile3..", version = "%prog 06052005", \
                          formatter = optparse.TitledHelpFormatter(),\
    description = '''This program will take the a set of hmm files (one hmm file one chromosome) generated by HMMTiling program and create One ChIP-enriched regions in UCSC bed format according to user specified criteria.  ''')

    parser.add_option("-o", dest="outfile",
                      help="output bed file")
    parser.add_option("--ip", dest="InterProbe", type="int",default=500,
                      help="maximum interpobe region, default 500")
    parser.add_option("--tp", dest="Tpeak", type="float",default=0.5,
                      help="peak of treatment, default 0.5,  \
                      at least one probe should have value greater than this peak")
    parser.add_option("--tc", dest="Tcutoff", type="float",default=0.00001,
                      help="boundary of the treatment enriched peak, default 0")
    parser.add_option("--pc", dest="pcutoff", type="int",default=2,
                      help="minum number of probes, default 2")
    parser.add_option("--cc", dest="Ccutoff", type="float",default=-5.0,
                      help="boundary of the control enriched peak, default -5.0")
    parser.add_option("--cp", dest="Cpeak", type="float",default=-15.0,
                      help="bottom of control, default -15.0, \
                      at least one probe should have value less than this bottom")
    parser.add_option("--extend", dest="extend", action="store_true",
                      help="add 25(probe length) to the position of the last probe")
    (options, args) = parser.parse_args()


    if len(sys.argv) <2:
        parser.print_help()
        sys.exit()

    main(options, args)

