#! /usr/bin/python

import sys, re, math, time
from numarray import *
sys.path.append( re.sub('[^\/]+$', 'SDK',sys.argv[0]))
from  MATRIX import Matrix             #SDK

class ProbeBehavior:
    '''self.probes={(ChipID,pos):(mean, stdev),...}'''
    def __init__(self, pbfilename):
        print >> sys.stderr, "Reading probe behavior file"
        self.probes={}
        filehandle = open(pbfilename)

        for x in filehandle:
            if not x[0].isdigit():
                continue
            x = x.split()
            self.probes[( int(x[0]), int(x[1]) )] =(float(x[2]), float(x[3]))



class Probe:        #data structure for a probe
    position = ''
    #probability in Non-enriched state
    pN = ''        #emission
    N=''
    #probability in Enriched state
    pY  = ''       #emission
    Y=''
    log_odds =''
    score =''


class MyMatrix(Matrix):
    ''' self.treatments = [probe1, probe2,...],self.controls = [probe1, probe2,...];
    '''

    def Read(self, pb, cols):
        self.treatments = []
        self.controls = []

        print >> sys.stderr, "Reading matrix file ", self.filename
        probes = {'treatments':[], 'controls':[]}
        for x in self.filehandle:
            if not x[0].isdigit():
                continue
            x = [float(i) for i in x.split()]

            #probe lines begin
            ChipID, pos = int(x[0]), int(x[1])
            (mean, stdev) = pb.probes[(ChipID,pos)]

            for group in ['treatments','controls']:
                pY = ''     # emission probability in Enriched State
                pN = ''     # emission probability in NonEnriched State
                Count = 0   #Number of real probes
                for column in cols[group]:
                    column = column + 2
                    if x[column] <= -999999:        #outlier probes
                        pass

                    #set the upper and lower boundaries
                    if ((x[column] - mean) / stdev > 3):
                        x[column] = mean + 3 * stdev
                    if ((x[column] - mean) / stdev < -2.5):
                        x[column]=mean -2.5 * stdev

                    z1 = (x[column] - mean) / stdev
                    z2 = (x[column] - mean - 2 * stdev) / stdev / 1.5
                    if not pY:          #the first probe
                        pY = self.getEmitP(z2)
                        pN = self.getEmitP(z1)
                    else:
                        pY += self.getEmitP(z2)
                        pN += self.getEmitP(z1)
                    Count += 1

                if not pY:              #outlier probes across all the experiments
                    probes[group].append(0)
                else:                   #real data
                    probe = Probe()
                    probe.position = pos
                    probe.pN = pN/Count
                    probe.pY  = pY/Count
                    probes[group].append(probe)

        self.treatments = probes['treatments']
        self.controls = probes['controls']
        self.num_probe =len(self.treatments)

    def getEmitP(self, value):
        ''' get the -0.5*x^2'''
        return -0.5 * math.pow(value, 2)

def main(options):
    HMMGap =500         #re-initialize HMM chain when distance between two probes are greater than HMMGap
    cols={}             #data columns index
    infile = MyMatrix(options.infile)
    pbfile = ProbeBehavior(options.pbfile)
    outfile = open(options.infile +".hmm","w")
    cols['treatments'] = [int(x) for x in options.treatments.split(",")]
    cols['controls'] = [int(x) for x in options.controls.split(",")]

    #check input column index and print the header line
    print >> outfile, "#"+infile.chr
    print >> outfile, "#", time.ctime(),"Running parametres:  " , " ".join(sys.argv)
    print >>outfile, "Pos\t",
    for x in ['treatments','controls']:
        Experiment = ""
        for i in cols[x]:
            if (i < 0 or i >= infile.num_exp):
                print >> sys.stderr, "Column can only be within [%d, %d]" % (0, infile.num_exp-1)
                sys.exit(0)
            if (not Experiment):
                Experiment = infile.celnames[i]
            else:
                Experiment += "/" + infile.celnames[i]
        print >> sys.stderr, "Check ", x, cols[x], "columns", "\n", Experiment
        print >>outfile, "%s\t" % (Experiment),
    print >>outfile, ""

    #Read the real data
    infile.Read(pbfile, cols)
    del pbfile

    #transition probability
    if options.numPeaks:
        prior = options.numPeaks  / infile.num_probe
    if options.rate:
        prior = options.rate
    StayP = math.log(1-prior)
    TranP = math.log(prior)

    for data in [infile.treatments,infile.controls]:

        print >> sys.stderr, "Performing backward HMM "
        #init
        nextProbe = infile.num_probe -1
        data[nextProbe].Y =Y = math.log(prior)
        data[nextProbe].N =N = math.log(1 - prior)

        for i in range(infile.num_probe -2, -1, -1):
            if (not data[i]):                       #outlier probes
                continue
            #re-initialize HMM chain
            if data[i].position- data[nextProbe].position < -HMMGap:
                Y = math.log(prior)
                N = math.log(1 - prior)

            # non-enriched
            temp1 = Y + TranP + data[nextProbe].pY
            temp2 = N + StayP + data[nextProbe].pN

            if (temp1 > temp2):
                data[i].N =  temp2 + math.log(math.exp(temp1 - temp2) + 1)
            else:
                data[i].N =  temp1 + math.log(math.exp(temp2 - temp1) + 1)

            # enriched
            temp1 = Y + StayP + data[nextProbe].pY
            temp2 = N + TranP + data[nextProbe].pN

            if (temp1 > temp2):
                data[i].Y = Y = temp2 + math.log(math.exp(temp1 - temp2) + 1)
            else:
                data[i].Y = Y = temp1 + math.log(math.exp(temp2 - temp1) + 1)
            N=data[i].N
            nextProbe = i

        print >> sys.stderr, "Performing forward HMM"
        #init
        Y = math.log(prior)
        N = math.log(1 - prior)
        preProbe = 0

        for i in range(1, len(data)):
            if (not data[i]):
                continue
            #re-initialize HMM chain
            if data[i].position- data[preProbe].position > HMMGap:
                Y = math.log(prior)
                N = math.log(1 - prior)

            temp1 = Y + TranP + data[i].pN
            temp2 = N + StayP + data[i].pN
            if (temp1 > temp2):
                pNCurrent = temp2 + math.log(math.exp(temp1 - temp2) + 1)
            else:
                pNCurrent = temp1 + math.log(math.exp(temp2 - temp1) + 1)
            temp1 = Y + StayP + data[i].pY
            temp2 = N + TranP + data[i].pY
            if (temp1 > temp2):
                Y = temp2 + math.log(math.exp(temp1 - temp2) + 1)
            else:
                Y = temp1 + math.log(math.exp(temp2 - temp1) + 1)
            N=pNCurrent

            preProbe = i

            # calculate marginal pb:
            temp1 = data[i].N + N
            temp2 = data[i].Y + Y
            data[i].log_odds = temp2 - temp1   # YY - NN
            if (data[i].log_odds > 0):
                data[i].score = math.exp(data[i].log_odds - math.log(math.exp(data[i].log_odds) + 1)) * 100
            else:
                data[i].score = math.exp(-math.log(math.exp(-data[i].log_odds) + 1)) * 100


    print >> sys.stderr, "Writing to disk ..."
    for i in range(1,infile.num_probe):
        if not infile.treatments[i] or not infile.controls[i]: # NA
            print >> outfile, "NA"
            continue
        print >> outfile, "%d\t%.1f\t%.1f" % (infile.treatments[i].position, infile.treatments[i].log_odds,\
                                            infile.controls[i].log_odds)

#read program options
if __name__ == '__main__':
    import  optparse
    parser = optparse.OptionParser("%prog -i MatrixFile [-p 300|-r 0.001] -t 3,4,5 -c 7,8,9 -b PBFile",\
                                   formatter = optparse.TitledHelpFormatter(),\
    description = '''This is the main program in this package. This program will take the matrix file generated from Cel2Matrix program to run HMM algorithm to get the log_odds of treatment and control groups for each probe. Users can run BedCall program afterwards to get the bed-format enriched region. It will add a .hmm suffix to the matrix file as the output hmm file. ''')

    parser.add_option("-i",  dest="infile",
                      help="input normalized probe file", metavar="FILE")
    parser.add_option("-p", dest="numPeaks", type="float",
                      help="number of peaks")
    parser.add_option("-r", dest="rate", type="float",
                      help="equals to peaks / #probes")
    parser.add_option("-t", dest="treatments",
                      help="treatment columns; 0-based, like 0,2,4")
    parser.add_option("-c",dest="controls",
                      help="control columns; 0-based, like 1,3,4")
    parser.add_option("-b", dest="pbfile",
                      help="probe behavior file")

    (options, args) = parser.parse_args()

    if options.numPeaks and options.rate:
        parser.error('only one can be specified: rate or numPeaks')

    if len(sys.argv) <2:
        parser.print_help()
        sys.exit()
    #end of option read

    main(options)


