#! /usr/bin/python


import sys,  math, re, time
sys.path.append( re.sub('[^\/]+$', 'SDK',sys.argv[0]))
from  MATRIX import Matrix             #SDK
from numarray import *


class MyMatrix(Matrix):

    def CutoffValue2Outlier(self,cols, percent):
        ''' make all probes above certain cutoff Outliers (-999999) '''
        cols = array(cols) +2        #make the ChipID and Pos at the beginning
        for col in cols:
            data = self.probes[:,col].copy()
            data.sort()
            data = compress(data > -999999,data)        #-999999 is outlier probes
            cutoff = data[ int(-1 * percent * data.size() )]
            print >> sys.stderr, str(percent) + " cutoff for " + self.celnames[col-2] + " is: " +str(cutoff)
            self.probes[:,col][ self.probes[:, col] >= cutoff] = -999999

def main(options, args):
    input =MyMatrix(options.infile)
    if not options.output:
        output = open(input.chr+ '.pb', 'w')
    else:
        output = open(options.output, "w")

    ParaList = range(input.num_exp)                    #cols not included in  probe behavior estimation
    if options.pcols:
        options.pcols = [int(i) for i in options.pcols.split(",")]
        print >> sys.stderr, "partial outlier columns with percentage cutoff  ", options.pcutoff
        for i in options.pcols :
            print >> sys.stderr, input.celnames[i ]
            ParaList.remove(i)

    if options.ecols:
        options.ecols =  [ int(i) for i in options.ecols.split(",")]
        print >> sys.stderr, "all probes in this experiments are included:"
        for i in options.ecols:
            print >> sys.stderr, input.celnames[i]
            ParaList.remove(i)

    print >> sys.stderr,  "completely excluded experiments : "
    for i,x in enumerate(ParaList):
        print >> sys.stderr, i, input.celnames[x]

    #read matrix file
    input.Read()

    #make all probes in pcols above certain cutoff Outliers (-999999)
    if options.pcols:
        input.CutoffValue2Outlier(options.pcols, options.pcutoff)

    #make all probes in ecols Outliers (-999999)
    if ParaList:
        input.CutoffValue2Outlier(ParaList,1)

    print >> output,"#"+input.chr
    print >> output, "#", time.ctime(),"Running parametres:  " , " ".join(sys.argv)
    print >> output, "ChipID\tPos\tMean\tStdev"
    for data in input.probes:
        temp = compress(data[2:] > -999999, data[2:]) # remove outlier probes
        if temp.size() > 1:
            print >> output, "%d\t%d\t%.1f\t%.1f" % (data[0], data[1], temp.mean(), temp.stddev() )
        else:       #outlier probe, make mean and stdev as -999999
            print >> output, "%d\t%d\t%.1f\t%.1f" % (data[0], data[1], -999999, -999999 )



#read program options
if __name__ == '__main__':
    import optparse

    parser = optparse.OptionParser(usage = "%prog -i Matrixfile -o PBfile --pcutoff  --pcols --ecols  ", \
                          version = "%prog 06102005", formatter = optparse.TitledHelpFormatter(), \
    description = '''
    This program will take the matrix file generated from Cel2Matrix program and create a probe behavior file. The default output is the chromosome name from the matrix file + ".pb" suffix like chr21.gb. You may specify the experiment index as partial outlier (you have to specify the percentage cutoff as well, the top percentage probes in this experiment are excluded in probe behavior estimation) or entire probe experiments (all probes in this experiment are included in estimation). So the rest of the experiments not mentioned are entire outliers (all probes in this experiment are excluded in estimation). The leftmost experiment is indexed as 0, the second left is indexed as 1 and so on.
    ''')
    parser.add_option("-i",  dest="infile",
                      help="Required. input matrix file", metavar="FILE")
    parser.add_option("-o", dest="output",
                      help='''Optional. output probe behavior file, default is chromosome name from matrix file + .pb suffix, like chr21.gb''')

    parser.add_option("--pcutoff", dest="pcutoff", type="float",
                      help="Optional. partial outlier cutoff (percentage like 0.005), the top 0.5% probes will not be included")

    parser.add_option("--pcols", dest="pcols",
                      help="Optional. partial outlier experiments index, like --pcols 0,1,2,3")
    parser.add_option("--ecols", dest="ecols",
                      help='''Optional. entire probe experimets index, all probes in this experiment are included in estimation,like --ecols 4,5,6''')
    (options, args) = parser.parse_args()

    if len(sys.argv) <2:
        parser.print_help()
        sys.exit()

    if (options.pcols and not options.pcutoff) or (not options.pcols and  options.pcutoff):
        parser.error("you should specify pcutoff and pcols together")

    if not options.pcols and not options.ecols:
        parser.error("you should specify pcols or ecols")

    main(options, args)
