#!/usr/bin/env python
# weili@jimmy.harvard.edu

from numarray import *
import sys, math, re, os
sys.path.append( re.sub('[^\/]+$', 'SDK/FileIO',sys.argv[0]))
import libAffx_Bruce_CELFile            #FileIO SDK

class MyCel(object):
    def __init__(self, filename):
        self.filename = filename
        self.mean=[]
        self.cel = libAffx_Bruce_CELFile.CELFile()
        self.cel.SetFileName(filename)
        if not self.cel.Exists():
            print >> sys.stderr, "Cannot read cel file ", filename

        print  >> sys.stderr, "Reading cel file ", filename
        self.cel.Read()
        self.Cols = self.cel.GetCols()
        self.Rows = self.cel.GetRows()
        self.size = self.Cols * self.Rows

    def GetMean(self):
        ''' un-sorted mean'''
        mean= zeros((self.Cols, self.Rows), Float32)
        for i in range(self.Cols):
            for j in range(self.Rows):
                mean[i,j] = self.cel.GetMean(i,j)

        mean.setshape(self.size)
        return mean

    def SetMean(self,norm):
        mean = self.GetMean()
        put(mean, mean.argsort(), norm)
        mean.setshape(self.Cols, self.Rows)
        for i in range(self.Cols):
            for j in range(self.Rows):
                self.cel.SetMean(i,j,mean[i,j])

    def Write(self,  path, format):
        newname = path +'/'+self.filename
        self.cel.SetFileName(newname)

        if format == 'xda':
            self.cel.WriteXDABCel()
        if format == 'text':
            self.cel.WriteTextCel()



def main(options, args):
    cels =[]
    for filename in args:
        cels.append(MyCel(filename))

    mean = zeros(cels[0].size, Float32)

    if not options.reference:                   #no reference cel file, self normalization
        for cel in cels:
            print >> sys.stderr, "Get the mean of ", cel.filename
            tmp = cel.GetMean()
            tmp.sort()
            mean = mean + tmp

        #take the average of the mean
        mean = mean / len(cels)
    else:                                       #reference cel file
        reference = MyCel(options.reference)
        tmp = reference.GetMean()
        tmp.sort()
        mean = tmp

    print >> sys.stderr, "Put the mean back and writing"
    for cel in cels:
        cel.SetMean(mean)
        cel.Write(options.outdir, options.outformat)



#read program options
if __name__ == '__main__':
    import sys, optparse

    parser = optparse.OptionParser(usage = "%prog  *.cel ..", version = "%prog 06102005",\
    formatter = optparse.TitledHelpFormatter(),\
    description = '''This program will take a set of input cel files and create quantile-normalized cel files. If you want to use the pre-computed probe behavior file in the HMM algorithm , you must choose a corresponding normalized cel file as reference in the quantile normalization.
    ''')
    parser.add_option("--reference",  dest="reference",
                      help="reference cel file for normalization. If you want to use the pre-computed probe behavior file in HMM, this option is required", metavar="FILE")

    parser.add_option("--outdir", dest="outdir", default='.',
                      help="Optional. Location to place quantile normalized cel files. Defaults to current directory")

    parser.add_option("--outformat",  dest="outformat",default = 'xda',
                      help="Optional. Sepcify the output file format. Options are text, xda binary (default)")
    (options, args) = parser.parse_args()

    if len(sys.argv) <2:
        parser.print_help()
        sys.exit()
    if not(options.outformat == 'xda' or options.outformat == 'text'):
        parser.error("outformat can only be either xda or text")

    dest = os.listdir(options.outdir)
    for filename in dest:
        if  args.count(filename):
            parser.error("outdir %s is not empty!" %(options.outdir))


    main(options, args)