/////////////////////////////////////////////////////////////////
//
// Copyright (C) 2004 Affymetrix, Inc.
//
// This library is free software; you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation; either version 2.1 of the License,
// or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
// for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this library; if not, write to the Free Software Foundation, Inc.,
// 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
//
/////////////////////////////////////////////////////////////////

#ifndef _AB_CelFile_h_
#define _AB_CelFile_h_

#include "CELFileData.h"
#include <vector>
#include <iostream>

//////////////////////////////////////////////////////////////////////
namespace affxcel 
{
	class CCELFileData;
}
namespace affxbruce
{
///////////////////////////////////////////////////////////////////////////////
///  affxbruce::CELFile
///  \brief Wrapper class of affxcel::CCELFileData for Bruce so as to implement the SWIG layer for perl
///
///  @see affxcel::CCELFileData
///////////////////////////////////////////////////////////////////////////////
class CELFile  
{
private:
	affxcel::CCELFileData *m_pCelFileData;

public:
	// Constructor/destructor
	CELFile();
	~CELFile();

	// Error.
	void SetError(char* str);
	const char* GetError() const;

	// File name parameters.
	void SetDataPath(const char *str);
	void SetLibPath(const char *str);
	void SetFileName(const char *str);
	const char* GetFileName() const;
	const char* GetFilePath() const;

	// Accessors for header information.
	const char* GetHeaderKey(char* key) const;
	int GetVersion() const;
	int GetCols() const;
	int GetRows() const;
	int GetNumCells() const;
	const char* GetHeader() const;
	const char* GetAlg() const;
	const char* GetParams() const;
	const char* GetDatHeader() const;
	const char* GetChipType() const;
	int GetCellMargin() const;
	unsigned long GetNumOutliers() const;
	unsigned long GetNumMasked() const;
	affxcel::GridCoordinatesType GetCorners() const;

	// Accessors for intensity information.
	float GetMean(int x, int y) const;
	float GetStdev(int x, int y) const;
	short GetPixels(int x, int y) const;
	int IsMasked(int x, int y) const;
	int IsOutlier(int x, int y) const;

	// For reading a file.
	void Close();
	long GetFileSize() const;
	int Exists();
	int ReadHeader();
	int Read();
	int IsXDACompatibleFile();
	int IsTranscriptomeBcelFile();
	int IsCompactCelFile();

	// For writing a new CEL file
	void Clear();
	int GetFileFormat();
	void SetFileFormat(int i);
	void SetAlgorithmName(const char *str);
	void AddAlgorithmParameter(const char *tag, const char *value);
	void SetDimensions(int rows, int cols);
	void SetChipType(const char *str);
	void SetMargin(int margin);
	void SetMean(int x, int y, float mean);
	void SetStdev(int x, int y, float stdev);
	void SetPixels(int x, int y, short pixels);
	void SetMasked(int x, int y, int masked);
	void SetOutlier(int x, int y, int outlier);
	int WriteTextCel();
	int WriteXDABCel();
	int WriteTranscriptomeBCel();
	int WriteCompactBCel();
	int TEXT_CEL();
	int XDA_BCEL();
	int TRANSCRIPTOME_BCEL();
	int COMPACT_BCEL();

  //
	affxcel::CCELFileData *GetThisPtr();
	void EnsureNotMmapped();

private:
	void CopyString(char*& pDest, const char* pSource) const;
	void AppendString(char*& pDest, const char* pSource) const;
};

//////////////////////////////////////////////////////////////////////

} // namespace

////////////////////////////////////////////////////////////////////


namespace std {

using namespace affxcel;

void CelFilePtrVector_pushref(std::vector<affxcel::CCELFileData *>&celptrvec,affxcel::CCELFileData *cel);
affxcel::CCELFileData *CelFilePtrVector_getref(std::vector<affxcel::CCELFileData *>&celptrvec,unsigned int i);

void CelFilePtrVector_print(std::vector<affxcel::CCELFileData *> &celptrvec);
}

#endif // _AB_CelFile_h_
