////////////////////////////////////////////////////////////////
//
// Copyright (C) 2004 Affymetrix, Inc.
//
// AFFYMETRIX, INC. LIMITED RESEARCH AND DEVELOPMENT LICENSE AGREEMENT
// 
// Affymetrix GeneChip Human Exon Array Data Analysis Software
// 		
// IMPORTANT - PLEASE READ CAREFULLY: This Limited Research and Development
// License Agreement (the Agreement) is a legally binding agreement between
// you (either an individual or an entity) and Affymetrix, Inc. for the
// version of the Affymetrix software product identified above provided in
// connection with this Agreement, which includes computer software (in
// source and/or object code form) and may include related media, printed
// materials and/or electronic documentation (the Software). This Agreement
// contains terms and conditions which are applicable to your use of the
// Software. By clicking "I Accept" below and/or accessing, downloading,
// viewing or using the Software, you are indicating that you agree to be
// bound by the following terms. If you do not agree to the terms of this
// Agreement, you are not licensed to and may not use the Software.
// 
// GRANT OF LICENSE.  Affymetrix grants you a royalty-free, non-exclusive,
// non-transferable, non-sublicensable, internal-use copyright and trade
// secret license to install and use the Software to analyze data generated
// using Affymetrix GeneChip human exon arrays, subject to all terms and
// conditions of this Agreement, the terms and conditions of the Technology
// Access Program or other agreement, if any, between you and Affymetrix (or
// one of its subsidiaries)  in connection with which the Software is
// provided, and the general terms and conditions of Affymetrix applicable to
// any product(s) to which the Software pertains. This license permits you to
// modify and make derivative works of any source code components, if any, of
// the Software and compile them into object code solely for your internal
// use consistent with the terms and conditions of this Agreement. Such
// license shall terminate within one hundred and eighty (180) calendar days
// from the day you first download or access the Software, or upon
// Affymetrix broad commercial launch, if any, of the GeneChip human exon
// array, or December 31, 2005, whichever is earlier. You agree to destroy,
// or return to Affymetrix upon Affymetrix request, all copies of the
// Software and cease using the Software upon termination of the license.
// 
// RESTRICTIONS ON SOFTWARE. You shall not: (i) modify or create derivative
// works of the Software, except as expressly permitted above for your
// internal use only; (ii) distribute, market, rent, lease or transfer to any
// third party the Software or any portion thereof; (iii) use the Software in
// any time sharing or rental arrangement; (iv) download, upload, or
// otherwise export or re-export the Software or any underlying information
// or technology (or direct product thereof) except in full compliance with
// all United States and other applicable laws and regulations; nor (v)
// remove any product identification, trademark, copyright or other notices.
// 
// INTELLECTUAL PROPERTY RIGHTS.  Subject to your limited license to use the
// Software as provided above, you agree that Affymetrix owns all copyright,
// patent, trademark, trade secret, and all other intellectual property
// rights in the Software. You hereby grant to Affymetrix a non-exclusive,
// worldwide, fully sublicensable, fully paid-up, royalty-free, irrevocable,
// perpetual license to any and all modifications or improvements you make to
// the Software code and any feedback or guidance you provide regarding the
// Software; the foregoing may be used, disclosed, disseminated and/or
// published by Affymetrix for any purpose, including incorporation into the
// Software or other software, without obligation of any kind to you.
// 	
// TRADEMARKS. No license, right, or interest in any Affymetrix trademark,
// trade name, service mark, or other designation is granted hereunder.
// 
// NO IMPLIED LICENSE. You acquire no rights or licenses, implied or
// otherwise, except those expressly and unambiguously granted herein.
// 
// NO PATENT LICENSE. Nothing contained herein shall be construed as
// conferring by implication, estoppel or otherwise any license or right
// under any patent, whether or not the exercise of any right herein granted
// necessarily employs an invention claimed by any existing or later issued
// patent.  Notwithstanding the preceding sentence, you shall have the right
// to exercise all rights expressly and unambiguously granted pursuant to
// this Agreement.
// 
// NO WARRANTY. The Software is provided AS IS and Affymetrix makes no
// warranty as to the Software. Affymetrix shall have no obligation under
// this Agreement to correct or notify you of any bugs, defects or errors in
// the Software or to otherwise support or maintain the Software. No warranty
// with respect to performance specifications of the Software is made. You
// acknowledge and agree that Affymetrix may modify or cease supporting the
// Software at any time with or without notice and the entire risk of using
// the Software is borne by you. You further acknowledge that Affymetrix has
// not promised or guaranteed to you that the Software will be commercially
// launched or made available to anyone in the future, that Affymetrix has no
// express or implied obligation to you to commercially launch or introduce
// the Software and that Affymetrix may elect not to introduce a product
// similar to or compatible with the Software. AFFYMETRIX DISCLAIMS ALL
// WARRANTIES, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE IMPLIED
// WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
// NON-INFRINGEMENT, RELATED TO THE SOFTWARE, ITS USE OR ANY INABILITY TO USE
// IT, THE RESULTS OF ITS USE AND THIS AGREEMENT.
// 
// U.S. GOVERNMENT END USERS. If you are an agency, department, or other
// entity of the United States government (Government), the use,
// duplication, reproduction, release, modification, disclosure or transfer
// of this product, or of any related documentation of any kind, including
// technical data, is restricted in accordance with Federal Acquisition
// Regulation (FAR) 12.212 for civilian agencies and Defense Federal
// Acquisition Regulation Supplement (DFARS) 227.7202 for military
// agencies. The Software is commercial computer software and commercial
// computer software documentation. The use of the Software by the Government
// is further restricted in accordance with the terms of this Agreement. The
// contractor/manufacturer is Affymetrix Inc., 3380 Central Expressway, Santa
// Clara, CA 95051.
// 
// LIMITATIONS OF LIABILITY AND DAMAGES. IN NO EVENT SHALL AFFYMETRIX BE
// LIABLE FOR ANY DAMAGES, WHETHER IN CONTRACT OR TORT (INCLUING NEGLIGENCE)
// OR UNDER ANY OTHER LEGAL THEORY, INCLUDING BUT NOT LIMITED TO DIRECT,
// CONSEQUENTIAL, SPECIAL, EXEMPLARY, INCIDENTAL AND INDIRECT DAMAGES,
// ARISING OUT OF OR IN CONNECTION WITH THIS AGREEMENT OR THE USE OR THE
// INABILITY TO USE THE SOFTWARE OR THE RESULTS OF USING THE SOFTWARE. THE
// LIABILITY OF AFFYMETRIX SHALL IN NO EVENT EXCEED FIFTY DOLLARS (U.S.
// $50.00). YOU ACKNOWLEDGE AND AGREE THAT THE RISKS OF LOSS HEREUNDER ARE
// REFLECTED IN THE PRICE OF THE SOFTWARE LICENSE, AND THAT THESE TERMS WOULD
// HAVE BEEN DIFFERENT IF THERE HAD BEEN A DIFFERENT ALLOCATION OF RISK.
// 
// NOT APPROVED FOR DIAGNOSTIC USE.  The Software is for research use only
// and not for use in diagnostic procedures.  The Software does not have the
// approval of the U.S. Food and Drug Administration (FDA), the Japanese
// Minister of Health, Labor and Welfare (MHLW) or other regulatory
// approval. No license is conveyed or implied for you to use, and you agree
// not to use, the Software or any results generated therefrom in any manner
// requiring FDA, MHLW or similar regulatory approval FDA.
// 
// GENERAL. This Agreement constitutes the entire agreement between you and
// Affymetrix with respect to the subject matter hereof and is the final,
// complete and exclusive statement of the terms of the Agreement,
// superseding all prior written and oral agreements, understandings and
// undertakings relating to the subject matter hereof. Modifications may be
// made only in writing and signed by an authorized corporate officer of
// Affymetrix. The waiver of any term or condition or any breach thereof
// shall not affect any other term or condition of this Agreement. You may
// not assign or transfer this Agreement without the prior written consent of
// Affymetrix. Without prejudice to any other rights, Affymetrix may
// terminate this Agreement if you fail to comply with the terms and
// conditions of this Agreement, the terms and conditions of the Technology
// Access Program or other agreement, if any, between you and Affymetrix (or
// one of its subsidiaries) in connection with which the Software is
// provided, or the general terms and conditions of Affymetrix applicable to
// any product(s) to which the Software pertains. In such event, you must
// destroy, or return to Affymetrix upon Affymetrix request, all copies of
// the Software and cease using the Software. This Agreement shall be
// governed by and construed according to the laws of California, without
// regard to conflict of law provisions. In the event that any provision of
// this Agreement or portion thereof is found to be illegal or unenforceable,
// the Agreement shall be construed without the unenforceable provision or
// portion thereof.
// 
// Should you have any questions concerning this license agreement, or if you
// desire to contact Affymetrix for any reason please write: Affymetrix,
// Inc., 3380 Central Expressway, Santa Clara, CA 95051, Attention: Legal
// Department; or phone 408.731.5000.
// 
// AS STATED ABOVE, USE OF THE SOFTWARE INDICATES YOUR ACCEPTANCE OF THIS
// AGREEMENT.
//
////////////////////////////////////////////////////////////////

// The file formats supported are:
//   text: text format of (x=int,y=int,mean=float,stdv=float,pixels=int)
//   bcel:   (mean=short,stdv=short,pixels=char)
//   xda:    (mean=float,stdv=float,pixels=short)
//   cbcel:  (mean=short) 

#include "CELFileData.h"
#include "FileIO.h"
#include "FileWriter.h"
#include <sys/types.h>
#include <sys/stat.h>
#include <istream>
#include <fstream>
#include <string.h>
#include <assert.h>
#include <algorithm>
#include <iostream>

using namespace affxcel;

//////////////////////////////////////////////////////////////////////

#define DELIMCHAR 0x14
#define MIN_CELLSTR 3

// JHG: This should have been written in hex from the beginning
//#define CELL_FILE_MAGIC_NUMBER 64
//#define CELL_FILE_VERSION_NUMBER 4
#define CELL_FILE_MAGIC_NUMBER   0x00000040 // 64
#define CELL_FILE_VERSION_NUMBER 0x00000004 // 4

// Transcriptome bcel
#define BCEL_HEADER_BYTES "BCEL\r\n\032\n"
#define BCEL_HEADER_LEN 8
#define STRUCT_SIZE_BCEL_HEADER_INFO 66
#define STRUCT_SIZE_FEATURE_DATA 5
#define STRUCT_SIZE_XY_PAIR 8
#define BCEL_CHUNK_LEN_SIZE 4
#define BCEL_CHUNK_NAME_SIZE 4
#define BCEL_CHUNK_NAME_SIZE0 5 /* for when we want a char string which can include a '\0' */
#define BCEL_CHUNK_CRC_SIZE 4
#define BCEL_CHUNK_HEAD "HEAD"
#define BCEL_CHUNK_DTHD "DTHD"
#define BCEL_CHUNK_ALGM "ALGM"
#define BCEL_CHUNK_ALPR "ALPR"
#define BCEL_CHUNK_INTY "INTY"
#define BCEL_CHUNK_MASK "MASK"
#define BCEL_CHUNK_OUTL "OUTL"
#define BCEL_CHUNK_TAIL "TAIL"

// Compact cel
#define CCEL_HEADER_BYTES "CCEL\r\n\064\n"
#define CCEL_HEADER_LEN 8


#ifdef WIN32
#define PATH_SEPARATOR '\\'
#define LINE_SEPARATOR "\n"
#else
#define PATH_SEPARATOR '/'
#define LINE_SEPARATOR "\r\n"
#endif

//jhg
//PAGE_SIZE
//PAGE_MASK
//PAGE_TRUNC(ptr)


// define if not available
#ifndef PAGE_SIZE
#define PAGE_SIZE (getpagesize())
#endif
#ifndef PAGE_MASK
#define PAGE_MASK ~(PAGE_SIZE-1)
#endif
#ifndef PAGE_TRUNC
#define PAGE_TRUNC(ptr) (ptr&(PAGE_MASK))
#endif 

//////////////////////////////////////////////////////////////////////

CCELFileHeaderData::CCELFileHeaderData()
{
	m_Magic = 0;
	m_Version = CELL_FILE_VERSION_NUMBER;
	m_nCols = 0;
	m_nRows = 0;
	m_nCells = 0;
	m_Margin = 0;
	m_nOutliers = 0;
	m_nMasked = 0;
	m_Header = "";
	m_Alg = "";
	m_Params = "";
	m_DatHeader = "";
	m_ChipType = "";
	m_CellGrid.upperleft.x = 0;
	m_CellGrid.upperleft.y = 0;
	m_CellGrid.upperright.x = 0;
	m_CellGrid.upperright.y = 0;
	m_CellGrid.lowerleft.x = 0;
	m_CellGrid.lowerleft.y = 0;
	m_CellGrid.lowerright.x = 0;
	m_CellGrid.lowerright.y = 0;
}

CCELFileHeaderData::~CCELFileHeaderData()
{
	Clear();
}

void CCELFileHeaderData::Clear()
{
	m_nRows = 0;
	m_nCols = 0;
	m_nCells = 0;
	if (m_ParameterIndices.size() > 0)
		m_ParameterIndices.clear();
	if (m_Parameters.size() > 0)
		m_Parameters.clear();
	m_Header = "";
	m_Alg = "";
	m_Params = "";
	m_DatHeader = "";
	m_ChipType = "";
	m_CellGrid.upperleft.x = 0;
	m_CellGrid.upperleft.y = 0;
	m_CellGrid.upperright.x = 0;
	m_CellGrid.upperright.y = 0;
	m_CellGrid.lowerleft.x = 0;
	m_CellGrid.lowerleft.y = 0;
	m_CellGrid.lowerright.x = 0;
	m_CellGrid.lowerright.y = 0;
	m_Margin = 1;
	m_nOutliers = 0;
	m_nMasked = 0;
	m_Magic = CELL_FILE_MAGIC_NUMBER;
	m_Version = CELL_FILE_VERSION_NUMBER;
}

std::string CCELFileHeaderData::GetHeader()
{
	SetDatHeader();
	char paramString[1024];
	sprintf(paramString, "Cols=%d\nRows=%d\nTotalX=%d\nTotalY=%d\nOffsetX=%d\nOffsetY=%d\n"
				"GridCornerUL=%ld %ld\nGridCornerUR=%ld %ld\nGridCornerLR=%ld %ld\nGridCornerLL=%ld %ld\n"
				"Axis-invertX=%d\nAxisInvertY=%d\nswapXY=%d\nDatHeader=%s"
				"\nAlgorithm=%s\nAlgorithmParameters=%s\n",
				m_nCols, m_nRows,
				m_nCols, m_nRows,
				0,0,
				m_CellGrid.upperleft.x, m_CellGrid.upperleft.y,
				m_CellGrid.upperright.x, m_CellGrid.upperright.y,
				m_CellGrid.lowerright.x, m_CellGrid.lowerright.y,
				m_CellGrid.lowerleft.x, m_CellGrid.lowerleft.y,
				0,0,0,
				m_DatHeader.c_str(),
				m_Alg.c_str(),
				GetAlgorithmParameters().c_str());

	m_Header = paramString;
	return m_Header;
}

void CCELFileHeaderData::SetDatHeader()
{
	if ((m_ChipType.length() > 0) && (m_DatHeader.length() == 0))
	{
		char scanInfo[512];
		sprintf(scanInfo, " %c %c %s.1sq %c %c %c %c %c %c %c %c %c ",
			DELIMCHAR,
			DELIMCHAR,
			m_ChipType.c_str(),
			DELIMCHAR,
			DELIMCHAR,
			DELIMCHAR,
			DELIMCHAR,
			DELIMCHAR,
			DELIMCHAR,
			DELIMCHAR,
			DELIMCHAR,
			DELIMCHAR);
		m_DatHeader = scanInfo;
	}
}

void CCELFileHeaderData::ParseDatHeader()
{
	const char *pInfo = strstr(m_Header.c_str(), "DatHeader=");
	if (pInfo != NULL)
	{
		const char *pEnd = strchr(pInfo, '\n');
		if (pEnd != NULL)
		{
			char *str = new char [pEnd - pInfo - 10 + 1];
			strncpy(str, pInfo + 10, pEnd - pInfo - 10);
			str[pEnd - pInfo - 10] = '\0';
			m_DatHeader = str;
			delete [] str;
		}
	}
}

void CCELFileHeaderData::ParseChipType()
{
	// Determine the array type from the header.
	// chip type is between second and third delimiters
	const char *pInfo = strchr(m_Header.c_str(), DELIMCHAR);
	if (pInfo != NULL)
	{
		pInfo = strchr(pInfo + 1, DELIMCHAR);
		if ((pInfo != NULL) && *(++pInfo))  
		{
			pInfo++;
			const char *pEnd = strchr(pInfo, '.');
			if (pEnd == NULL)  
			{
				pEnd = strchr(pInfo, DELIMCHAR);
				if (pEnd != NULL)
					pEnd--;
			}
			if (pEnd != NULL)
			{
				char *str = new char [pEnd - pInfo + 1];
				strncpy(str, pInfo, pEnd - pInfo);
				str[pEnd - pInfo] = '\0';
				m_ChipType = str;
				delete [] str;
			}
		}
	}
}

void CCELFileHeaderData::ParseAlgorithmParameters()
{
	if (m_Params.length() > 0)
	{
		int iStart = 0;
		int iEnd = 0;
		int iLen = (int) m_Params.length() - 1;
		std::string tag = "";
		std::string value = "";
		for (int i = 0; i < iLen; i++)
		{
			if ((m_Params.at(i) == ':') || (m_Params.at(i) == '='))
			{
				tag = m_Params.substr(iStart, (iEnd - iStart + 1));
				iStart = i + 1;
				iEnd = iStart;
			}
			else if ((m_Params.at(i) == ';') || (m_Params.at(i) == ' '))
			{
				value = m_Params.substr(iStart, (iEnd - iStart + 1));
				AddAlgorithmParameter(tag, value);
				iStart = i + 1;
				iEnd = iStart;
			}
			else
				iEnd = i;
		}
		value = m_Params.substr(iStart, (iEnd - iStart + 2));
		AddAlgorithmParameter(tag, value);
	}
}

void CCELFileHeaderData::AddAlgorithmParameter(const char *tag, const char *value)
{
	assert(tag != NULL);
	assert(value != NULL);

	std::string strTag = tag;
	std::string strValue = value;

	AddAlgorithmParameter(strTag, strValue);
}

void CCELFileHeaderData::AddAlgorithmParameter(std::string& tag, std::string& value)
{
	assert(tag != "");
	assert(value != "");

	std::map<std::string, std::string>::iterator pos = m_Parameters.find(tag);
	if (pos == m_Parameters.end())
	{
		int index = (int) m_Parameters.size() + 1;
		m_ParameterIndices.insert(std::make_pair(index, tag));
		m_Parameters.insert(std::make_pair(tag, value));
	}
}

void CCELFileHeaderData::SetAlgorithmParameter(const char *tag, const char *value)
{
	assert(tag != NULL);
	assert(value != NULL);

	std::string strTag = tag;
	std::string strValue = value;
	std::map<std::string, std::string>::iterator pos = m_Parameters.find(strTag);
	if (pos != m_Parameters.end())
		pos->second == strValue;
}

std::string CCELFileHeaderData::GetAlgorithmParameter(const char *tag)
{
	assert(tag != NULL);

	std::string strTag = tag;
	std::string strValue = "";
	std::map<std::string, std::string>::iterator pos = m_Parameters.find(strTag);
	if (pos != m_Parameters.end())
		strValue = pos->second;
	return strValue;
}

std::string CCELFileHeaderData::GetAlgorithmParameters()
{
	std::string strBuffer = "";
	std::string strTag = "";
	std::string strValue = "";
	int iCount = 0;
	std::map<int, std::string>::iterator posIndex;
	std::map<std::string, std::string>::iterator pos;
	for (int i = 0; i < (int) m_ParameterIndices.size(); i++)
	{
		posIndex = m_ParameterIndices.find(i + 1);
		if (posIndex != m_ParameterIndices.end())
		{
			strTag = posIndex->second;
			pos = m_Parameters.find(strTag);
			if (pos != m_Parameters.end())
			{
				strValue = pos->second;
				if (iCount > 0) strBuffer.append(";"); 
				strBuffer.append(strTag);
				strBuffer.append(":"), 
				strBuffer.append(strValue);
				iCount++;
			}
		}
	}
	return strBuffer;
}

void CCELFileHeaderData::SetGrid(GridCoordinatesType grid)
{
	m_CellGrid.upperleft.x = grid.upperleft.x;
	m_CellGrid.upperleft.y = grid.upperleft.y;
	m_CellGrid.upperright.x = grid.upperright.x;
	m_CellGrid.upperright.y = grid.upperright.y;
	m_CellGrid.lowerleft.x = grid.lowerleft.x;
	m_CellGrid.lowerleft.y = grid.lowerleft.y;
	m_CellGrid.lowerright.x = grid.lowerright.x;
	m_CellGrid.lowerright.y = grid.lowerright.y;
}


void CCELFileHeaderData::ParseCorners()
{
    // get grid coordinates and size from the header
	if (m_Header.length() > 0)
	{
		const char* pch = strstr(m_Header.c_str(), "GridCorner");
		int ulx,uly;
		int urx,ury;
		int llx,lly;
		int lrx,lry;
		if ((pch != NULL) &&
			(sscanf(pch, "GridCornerUL=%d %d GridCornerUR=%d %d GridCornerLR=%d %d GridCornerLL=%d %d",
						&ulx, &uly,
						&urx, &ury,
						&lrx, &lry,
						&llx, &lly) == 8))
		{
			m_CellGrid.upperleft.x = ulx;
			m_CellGrid.upperleft.y = uly;

			m_CellGrid.upperright.x = urx;
			m_CellGrid.upperright.y = ury;

			m_CellGrid.lowerright.x = lrx;
			m_CellGrid.lowerright.y = lry;

			m_CellGrid.lowerleft.x = llx;
			m_CellGrid.lowerleft.y = lly;
		}
	}
}

//////////////////////////////////////////////////////////////////////

void CCELFileData::SetDataPath(const char *str)
{ 
	assert(str != NULL);
	int iLen = (int) strlen(str);
	assert(iLen > 0);
	m_DataPath = str; 
}

void CCELFileData::SetLibPath(const char *str) 
{ 
	assert(str != NULL);
	int iLen = (int) strlen(str);
	assert(iLen > 0);
	m_LibPath = str; 
}

void CCELFileData::SetFileName(const char *str)
{
	assert(str != NULL);
	int iLen = (int) strlen(str);
	assert(iLen > 0);
	char* p = strrchr(str, PATH_SEPARATOR);
	m_FullPathName = str;
	if (p != NULL)
	{
		m_FileName = p + 1;
		m_DataPath = str;
		m_DataPath[(int)(p - str)] = '\0';
	}
	else
		m_FileName = str;
}

//////////////////////////////////////////////////////////////////////

bool CCELFileData::Exists()
{
	assert(m_FullPathName != "");

	// Find the file stats.
	struct stat st;
	return ((stat(m_FullPathName.c_str(), &st) == 0)? true: false);
}

//////////////////////////////////////////////////////////////////////

long CCELFileData::GetFileSize()
{
	assert(m_FullPathName != "");

	long lSize = 0;
	struct stat st;
	if (stat(m_FullPathName.c_str(), &st) == 0)
	{
		lSize = st.st_size;
	}
	return lSize;
}

//////////////////////////////////////////////////////////////////////

bool CCELFileData::Open(bool bReadHeaderOnly)
{
	bool retVal = false;

	// First close the file.
	Close();

	DetermineFileFormat();
	if (IsXDACompatibleFile())
		retVal = ReadXDABCel(bReadHeaderOnly);
	else if (IsTranscriptomeBcelFile())
		retVal = ReadTranscriptomeBCel(bReadHeaderOnly);
	else if (IsCompactCelFile())
		retVal = ReadCompactBCel(bReadHeaderOnly);
	else
		retVal = ReadTextCel(bReadHeaderOnly);

	return retVal;
}
//////////////////////////////////////////////////////////////////////

bool CCELFileData::ReadXDABCel(bool bReadHeaderOnly)
{
	bool retVal = false;

	// Open the file.
	std::ifstream instr;
	instr.open(m_FullPathName.c_str(), std::ios::in | std::ios::binary);

	// Check if open
	if (!instr)
	{
		SetError("Unable to open the file.");
		return false;
	}

	Clear();

	// Read the header
	int iHeaderBytes = 0;
	int32_t magic;

	int32_t version;
	int32_t nSubGrids;
	char *sval=NULL;

	// Read the magic number.
	READ_INT(instr, magic);
	iHeaderBytes += INT_SIZE;
	m_HeaderData.SetMagic(magic);

	// Check if new type.
	if (!(magic==CELL_FILE_MAGIC_NUMBER))
	{
		SetError("The file does not appear to be the correct format.");
		return false;
	}

	// Read the version
	ReadInt32_I(instr, version);
	iHeaderBytes += INT_SIZE;
	m_HeaderData.SetVersion(version);

	// Read the dimensions of the array
	int32_t iValue = 0;
	ULONG_T ulValue = 0;
	ReadInt32_I(instr, iValue);
	iHeaderBytes += INT_SIZE;
	m_HeaderData.SetRows(iValue);
	ReadInt32_I(instr, iValue);
	iHeaderBytes += INT_SIZE;
	m_HeaderData.SetCols(iValue);
	ReadInt32_I(instr, iValue);
	iHeaderBytes += INT_SIZE;
	m_HeaderData.SetCells(iValue);

	// Read the other members.
	READ_STRING(instr, sval);
	iHeaderBytes += INT_SIZE;
	iHeaderBytes += ((int) strlen(sval));
	m_HeaderData.SetHeader(sval);
	delete[] sval;
	sval = NULL;
	READ_STRING(instr, sval);
	iHeaderBytes += INT_SIZE;
	iHeaderBytes += ((int) strlen(sval));
	m_HeaderData.SetAlg(sval);
	delete[] sval;
	sval = NULL;
	READ_STRING(instr, sval);
	iHeaderBytes += INT_SIZE;
	iHeaderBytes += ((int) strlen(sval));
	m_HeaderData.SetParams(sval);
	delete[] sval;
	sval = NULL;
	ReadInt32_I(instr, iValue);
	iHeaderBytes += INT_SIZE;
	m_HeaderData.SetMargin(iValue);
	ReadUInt32_I(instr, ulValue);
	iHeaderBytes += ULONG_SIZE;
	m_HeaderData.SetOutliers(ulValue);
	ReadUInt32_I(instr, ulValue);
	iHeaderBytes += ULONG_SIZE;
	m_HeaderData.SetMasked(ulValue);
	ReadInt32_I(instr, nSubGrids);
	iHeaderBytes += INT_SIZE;

	// Set the chip type and DatHeader
	m_HeaderData.ParseChipType();
	m_HeaderData.ParseDatHeader();

	// Parse algorithm parameters into map
	m_HeaderData.ParseAlgorithmParameters();

	// Set grid coordinates
	m_HeaderData.ParseCorners();

	instr.close();

	// Read the remaining data.
	if (bReadHeaderOnly)
		return true;

	// Memory map file
#ifdef WIN32
	SYSTEM_INFO info;
	GetSystemInfo(&info);
	m_hFile = CreateFile(m_FullPathName.c_str(), GENERIC_READ, FILE_SHARE_READ,
			NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL/*FILE_FLAG_RANDOM_ACCESS*/);
	if (m_hFile == INVALID_HANDLE_VALUE)
	{
		SetError("Failed to open the file for memory mapping.");
		return false;
	}
	m_hFileMap = CreateFileMapping(m_hFile, NULL, PAGE_READONLY, 0, 0, NULL);
	if (m_hFileMap != NULL)
	{
		m_lpFileMap = MapViewOfFile(m_hFileMap, FILE_MAP_READ, 0, 0, 0);
		if (m_lpFileMap == NULL)
		{
			Close();
			SetError("Unable to map view for the memory map file.");
			return false;
		}
		else
		{
			m_lpData = (char *)(m_lpFileMap) + iHeaderBytes;
		}
	}
#else
	long lFileSize = GetFileSize();
	char* szBuffer = new char[iHeaderBytes + 1];
	m_File = fopen(m_FullPathName.c_str(), "r");
	if (m_File == NULL)
	{
		SetError("Failed to open the file for memory mapping.");
		return false;
	}
	fread(szBuffer, iHeaderBytes, 1, m_File);
	delete [] szBuffer;
  
  size_t jhg_page_size=PAGE_SIZE;
  size_t jhg_page_mask=PAGE_MASK;
  size_t jhg_page_trunc1=PAGE_TRUNC(1);
  size_t jhg_page_trunc2=PAGE_TRUNC(0xffffff);

    size_t cellf_page_start = PAGE_TRUNC(ftell(m_File));
    size_t cellf_page_offset = ftell(m_File) - cellf_page_start;
    m_MapLen = lFileSize - cellf_page_start;
    m_lpFileMap = mmap(NULL, m_MapLen, PROT_READ, MAP_SHARED, fileno(m_File), cellf_page_start);
    if (m_lpFileMap == MAP_FAILED)
	{
		Close();
		SetError("Unable to map view for the memory map file.");
		return false;
	}
	else
	{
		m_lpData = (char *)(m_lpFileMap) + cellf_page_offset;
	}
	if (m_File != NULL) 
	{
		fclose(m_File);
		m_File = NULL;
	}
#endif

	short x=0;
	short y=0;
	int iCell;

	// "Read" the Mean data
	m_pEntries = (CELFileEntryType*)m_lpData;

	// Read the mask data
	int iOffset = m_HeaderData.GetCells() * (FLOAT_SIZE + FLOAT_SIZE + SHORT_SIZE);
	if (m_bReadMaskedCells)
	{
		for (iCell = 0; iCell < (int) m_HeaderData.GetMasked(); iCell++)
		{
			// Read the coordinate.
			x = GetShort((short*) (m_lpData + iOffset + iCell * 2 * SHORT_SIZE), m_FileFormat);
			y = GetShort((short*) (m_lpData + iOffset + iCell * 2 * SHORT_SIZE + SHORT_SIZE), m_FileFormat);
			m_MaskedCells.insert(std::make_pair(y * m_HeaderData.GetCols() + x, true));
		}
	}
		
	// Read the outlier data
	iOffset += m_HeaderData.GetMasked() * 2 * SHORT_SIZE;
	if (m_bReadOutliers)
	{
		for (iCell = 0; iCell < (int) m_HeaderData.GetOutliers(); iCell++)
		{
			// Read the coordinate.
			x = GetShort((short*) (m_lpData + iOffset + iCell * 2 * SHORT_SIZE), m_FileFormat);
			y = GetShort((short*) (m_lpData + iOffset + iCell * 2 * SHORT_SIZE + SHORT_SIZE), m_FileFormat);
			m_Outliers.insert(std::make_pair(y * m_HeaderData.GetCols() + x, true));
		}
	}
	retVal = true;
	return retVal;
}
//////////////////////////////////////////////////////////////////////

bool CCELFileData::ReadTranscriptomeBCel(bool bReadHeaderOnly)
{
	bool retVal = false;

	// Open the file.
	std::ifstream instr;
	instr.open(m_FullPathName.c_str(), std::ios::in | std::ios::binary);

	// Check if open
	if (!instr)
	{
		SetError("Unable to open the file.");
		return false;
	}

	Clear();

	// Read the header
	int iHeaderBytes = 0;
	float version = 0;
	ULONG_T ulValue = 0;
	uint32_t totalx = 0;
	uint32_t totaly = 0;
	uint32_t offsetx = 0;
	uint32_t offsety = 0;
	GridCoordinatesType grid;
	unsigned short axisinvertx = 0;
	unsigned short axisinverty = 0;
	unsigned short swapxy = 0;
	std::string scanInfo;
	std::string sval;

	// Read the magic number.
	ReadFixedString(instr, sval, BCEL_HEADER_LEN);
	iHeaderBytes += BCEL_HEADER_LEN;

	// Check if new type.
	if (strncmp(sval.c_str(), BCEL_HEADER_BYTES, BCEL_HEADER_LEN) != 0)
	{
		SetError("The file does not appear to be the correct format.");
		return false;
	}

	// Read the version
	ReadFloat_N(instr, version);
	iHeaderBytes += FLOAT_SIZE;
	m_HeaderData.SetVersion((int)version);

	// Read header information
	ReadUInt32_N(instr, ulValue);							// header struct size
	iHeaderBytes += (2 * ULONG_SIZE + ulValue + BCEL_CHUNK_NAME_SIZE);
	ReadFixedString(instr, sval, BCEL_CHUNK_NAME_SIZE);
	ReadFloat_N(instr, version);
	ReadUInt32_N(instr, ulValue);							// columns
	m_HeaderData.SetCols(ulValue);
	ReadUInt32_N(instr, ulValue);							// rows
	m_HeaderData.SetRows(ulValue);
	ReadUInt32_N(instr, totalx);							// totalx
	ReadUInt32_N(instr, totaly);							// totaly
	ReadUInt32_N(instr, offsetx);							// offsetx
	ReadUInt32_N(instr, offsety);							// offsety
	ReadUInt32_N(instr, ulValue);							// GridCornerUL.x
	grid.upperleft.x = ulValue;
	ReadUInt32_N(instr, ulValue);							// GridCornerUL.y
	grid.upperleft.y = ulValue;
	ReadUInt32_N(instr, ulValue);							// GridCornerUR.x
	grid.upperright.x = ulValue;
	ReadUInt32_N(instr, ulValue);							// GridCornerUR.y
	grid.upperright.y = ulValue;
	ReadUInt32_N(instr, ulValue);							// GridCornerLL.x
	grid.lowerleft.x = ulValue;
	ReadUInt32_N(instr, ulValue);							// GridCornerLL.y
	grid.lowerleft.y = ulValue;
	ReadUInt32_N(instr, ulValue);							// GridCornerLR.x
	grid.lowerright.x = ulValue;
	ReadUInt32_N(instr, ulValue);							// GridCornerLR.y
	grid.lowerright.y = ulValue;
	m_HeaderData.SetGrid(grid);
	ReadUInt16_N(instr, axisinvertx);						// Axis-invertX
	ReadUInt16_N(instr, axisinverty);						// AxisinvertY
	ReadUInt16_N(instr, swapxy);							// swapXY
	ReadUInt32_N(instr, ulValue);							

	// Read DatHeader
	ReadUInt32_N(instr, ulValue);							// DatHeader length
	iHeaderBytes += (2 * ULONG_SIZE + ulValue + BCEL_CHUNK_NAME_SIZE);
	ReadFixedString(instr, sval, BCEL_CHUNK_NAME_SIZE);
	ReadFixedString(instr, scanInfo, ulValue);
	if (scanInfo[ulValue - 1] == '\r') scanInfo[ulValue - 1] = '\0';
	m_HeaderData.SetDatHeader(scanInfo.c_str());
	ReadUInt32_N(instr, ulValue);							

	// Read Algorithm
	ReadUInt32_N(instr, ulValue);							// Algorithm length
	iHeaderBytes += (2 * ULONG_SIZE + ulValue + BCEL_CHUNK_NAME_SIZE);
	ReadFixedString(instr, sval, BCEL_CHUNK_NAME_SIZE);
	ReadFixedString(instr, sval, ulValue);
	if (sval[ulValue - 1] == '\r') sval[ulValue - 1] = '\0';
	m_HeaderData.SetAlg(sval.c_str());
	ReadUInt32_N(instr, ulValue);							

	// Read AlgorithmParameters
	ReadUInt32_N(instr, ulValue);							// AlgorithmParameters length
	iHeaderBytes += (2 * ULONG_SIZE + ulValue + BCEL_CHUNK_NAME_SIZE);
	ReadFixedString(instr, sval, BCEL_CHUNK_NAME_SIZE);
	ReadFixedString(instr, sval, ulValue);
	if (sval[ulValue - 1] == '\r') sval[ulValue - 1] = '\0';
	m_HeaderData.SetParams(sval.c_str());
	ReadUInt32_N(instr, ulValue);							

	char paramString[1024];
	sprintf(paramString, 
          "Cols=%d\nRows=%d\n"
          "TotalX=%u\nTotalY=%u\nOffsetX=%u\nOffsetY=%u\n"
          "GridCornerUL=%ld %ld\nGridCornerUR=%ld %ld\n"
          "GridCornerLR=%ld %ld\nGridCornerLL=%ld %ld\n"
          "Axis-invertX=%d\nAxisInvertY=%d\nswapXY=%d\nDatHeader=%s\n"
          "Algorithm=%s\nAlgorithmParameters=%s\n",
          m_HeaderData.GetCols(),m_HeaderData.GetRows(),
          totalx,totaly,offsetx,offsety,
          grid.upperleft.x,grid.upperleft.y,
          grid.upperright.x,grid.upperright.y,
          grid.lowerright.x,grid.lowerright.y,
          grid.lowerleft.x,grid.lowerleft.y,
          axisinvertx,axisinverty,swapxy,
          scanInfo.c_str(),
          m_HeaderData.GetAlg().c_str(),
          m_HeaderData.GetParams().c_str());
	m_HeaderData.SetHeader(paramString);

	// Read cell size
	ReadUInt32_N(instr, ulValue);
	int iOffset = ulValue;
	m_HeaderData.SetCells(ulValue / STRUCT_SIZE_FEATURE_DATA);
	ReadFixedString(instr, sval, BCEL_CHUNK_NAME_SIZE);
	iHeaderBytes += (ULONG_SIZE + BCEL_CHUNK_NAME_SIZE);

	// Set the chip type
	m_HeaderData.ParseChipType();

	// Parse algorithm parameters into map
	m_HeaderData.ParseAlgorithmParameters();

	// Set grid coordinates
	m_HeaderData.ParseCorners();

	// Read the remaining data.
	if (bReadHeaderOnly)
		return true;

	instr.close();

	// Memory map file
#ifdef WIN32
	SYSTEM_INFO info;
	GetSystemInfo(&info);
	m_hFile = CreateFile(m_FullPathName.c_str(), GENERIC_READ, FILE_SHARE_READ,
			NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL/*FILE_FLAG_RANDOM_ACCESS*/);
	if (m_hFile == INVALID_HANDLE_VALUE)
	{
		SetError("Failed to open the file for memory mapping.");
		return false;
	}
	m_hFileMap = CreateFileMapping(m_hFile, NULL, PAGE_READONLY, 0, 0, NULL);
	if (m_hFileMap != NULL)
	{
		m_lpFileMap = MapViewOfFile(m_hFileMap, FILE_MAP_READ, 0, 0, 0);
		if (m_lpFileMap == NULL)
		{
			Close();
			SetError("Unable to map view for the memory map file.");
			return false;
		}
		else
		{
			m_lpData = (char *)(m_lpFileMap) + iHeaderBytes;
		}
	}
#else
	long lFileSize = GetFileSize();
	char* szBuffer = new char[iHeaderBytes + 1];
	m_File = fopen(m_FullPathName.c_str(), "r");
	if (m_File == NULL)
	{
		SetError("Failed to open the file for memory mapping.");
		return false;
	}
	fread(szBuffer, iHeaderBytes, 1, m_File);
	delete [] szBuffer;

    size_t cellf_page_start = PAGE_TRUNC(ftell(m_File));
    size_t cellf_page_offset = ftell(m_File) - cellf_page_start;
    m_MapLen = lFileSize - cellf_page_start;
    m_lpFileMap = mmap(NULL, m_MapLen, PROT_READ, MAP_SHARED, fileno(m_File), cellf_page_start);
    if (m_lpFileMap == MAP_FAILED)
	{
		Close();
		SetError("Unable to map view for the memory map file.");
		return false;
	}
	else
	{
		m_lpData = (char *)(m_lpFileMap) + cellf_page_offset;
	}
	if (m_File != NULL) 
	{
		fclose(m_File);
		m_File = NULL;
	}
#endif

	// Read the Mean data
	ULONG_T x = 0;
	ULONG_T y = 0;
	int iCell;
	m_pTransciptomeEntries = (CELFileTranscriptomeEntryType*) m_lpData;

	// Read mask size
	iOffset = m_HeaderData.GetCells() * STRUCT_SIZE_FEATURE_DATA + ULONG_SIZE;
	ulValue = MmGetUInt32_N((ULONG_T*)(m_lpData + iOffset));
	m_HeaderData.SetMasked(ulValue / STRUCT_SIZE_XY_PAIR);
	iOffset += (ULONG_SIZE + BCEL_CHUNK_NAME_SIZE);

	// Read the mask data
	if (m_bReadMaskedCells)
	{
		for (iCell = 0; iCell < (int)m_HeaderData.GetMasked(); iCell++)
		{
			// Read the coordinate.
			x = MmGetUInt32_N((ULONG_T*) (m_lpData + iOffset + iCell * 2 * ULONG_SIZE));
			y = MmGetUInt32_N((ULONG_T*) (m_lpData + iOffset + iCell * 2 * ULONG_SIZE + ULONG_SIZE));
			m_MaskedCells.insert(std::make_pair(y * m_HeaderData.GetCols() + x, true));
		}
	}
	iOffset += (m_HeaderData.GetMasked() * STRUCT_SIZE_XY_PAIR + ULONG_SIZE);

	// Read outlier size
	ulValue = MmGetUInt32_N((ULONG_T*)(m_lpData + iOffset));
	m_HeaderData.SetOutliers(ulValue / STRUCT_SIZE_XY_PAIR);
	iOffset += (ULONG_SIZE + BCEL_CHUNK_NAME_SIZE);

	// Read the outlier data
	if (m_bReadOutliers)
	{
		for (iCell = 0; iCell < (int)m_HeaderData.GetOutliers(); iCell++)
		{
			// Read the coordinate.
			x = MmGetUInt32_N((ULONG_T*) (m_lpData + iOffset + iCell * 2 * ULONG_SIZE));
			y = MmGetUInt32_N((ULONG_T*) (m_lpData + iOffset + iCell * 2 * ULONG_SIZE + ULONG_SIZE));
			m_Outliers.insert(std::make_pair(y * m_HeaderData.GetCols() + x, true));
		}
	}

	retVal = true;
	return retVal;
}

//////////////////////////////////////////////////////////////////////

bool CCELFileData::ReadCompactBCel(bool bReadHeaderOnly)
{
	bool retVal = false;

	// Open the file.
	std::ifstream instr;
	instr.open(m_FullPathName.c_str(), std::ios::in | std::ios::binary);

	// Check if open
	if (!instr)
	{
		SetError("Unable to open the file.");
		return false;
	}

	Clear();

	// Read the header
	int iHeaderBytes = 0;
	int32_t version;
	int32_t nSubGrids;
	char *sval = NULL;

	// Read the magic number.
	std::string magic;
	ReadFixedString(instr, magic, CCEL_HEADER_LEN);
	iHeaderBytes += CCEL_HEADER_LEN;

	// Check if new type.
	if (strncmp(magic.c_str(), CCEL_HEADER_BYTES, CCEL_HEADER_LEN) != 0)
	{
		SetError("The file does not appear to be the correct format.");
		return false;
	}

	// Read the version
	ReadInt32_I(instr, version);
	iHeaderBytes += INT_SIZE;
	m_HeaderData.SetVersion(version);

	// Read the dimensions of the array
	int32_t iValue = 0;
	ReadInt32_I(instr, iValue);
	iHeaderBytes += INT_SIZE;
	m_HeaderData.SetRows(iValue);
	ReadInt32_I(instr, iValue);
	iHeaderBytes += INT_SIZE;
	m_HeaderData.SetCols(iValue);
	ReadInt32_I(instr, iValue);
	iHeaderBytes += INT_SIZE;
	m_HeaderData.SetCells(iValue);

	// Read the other members.
	READ_STRING(instr, sval);
	iHeaderBytes += INT_SIZE;
	iHeaderBytes += ((int) strlen(sval));
	m_HeaderData.SetHeader(sval);
	delete[] sval;
	sval = NULL;
	READ_STRING(instr, sval);
	iHeaderBytes += INT_SIZE;
	iHeaderBytes += ((int) strlen(sval));
	m_HeaderData.SetAlg(sval);
	delete[] sval;
	sval = NULL;
	READ_STRING(instr, sval);
	iHeaderBytes += INT_SIZE;
	iHeaderBytes += ((int) strlen(sval));
	m_HeaderData.SetParams(sval);
	delete[] sval;
	sval = NULL;
	ReadInt32_I(instr, iValue);
	iHeaderBytes += INT_SIZE;
	m_HeaderData.SetMargin(iValue);
	ReadInt32_I(instr, nSubGrids);
	iHeaderBytes += INT_SIZE;

	// Set the chip type and DatHeader
	m_HeaderData.ParseChipType();
	m_HeaderData.ParseDatHeader();

	// Parse algorithm parameters into map
	m_HeaderData.ParseAlgorithmParameters();

	// Set grid coordinates
	m_HeaderData.ParseCorners();

	instr.close();

	// Read the remaining data.
	if (bReadHeaderOnly)
		return true;

	// Memory map file
#ifdef WIN32
	SYSTEM_INFO info;
	GetSystemInfo(&info);
	m_hFile = CreateFile(m_FullPathName.c_str(), GENERIC_READ, FILE_SHARE_READ,
			NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL/*FILE_FLAG_RANDOM_ACCESS*/);
	if (m_hFile == INVALID_HANDLE_VALUE)
	{
		SetError("Failed to open the file for memory mapping.");
		return false;
	}
	m_hFileMap = CreateFileMapping(m_hFile, NULL, PAGE_READONLY, 0, 0, NULL);
	if (m_hFileMap != NULL)
	{
		m_lpFileMap = MapViewOfFile(m_hFileMap, FILE_MAP_READ, 0, 0, 0);
		if (m_lpFileMap == NULL)
		{
			Close();
			SetError("Unable to map view for the memory map file.");
			return false;
		}
		else
		{
			m_lpData = (char *)(m_lpFileMap) + iHeaderBytes;
		}
	}
#else
	long lFileSize = GetFileSize();
	char* szBuffer = new char[iHeaderBytes + 1];
	m_File = fopen(m_FullPathName.c_str(), "r");
	if (m_File == NULL)
	{
		SetError("Failed to open the file for memory mapping.");
		return false;
	}
	fread(szBuffer, iHeaderBytes, 1, m_File);
	delete [] szBuffer;

    size_t cellf_page_start = PAGE_TRUNC(ftell(m_File));
    size_t cellf_page_offset = ftell(m_File) - cellf_page_start;
    m_MapLen = lFileSize - cellf_page_start;
    m_lpFileMap = mmap(NULL, m_MapLen, PROT_READ, MAP_SHARED, fileno(m_File), cellf_page_start);
    if (m_lpFileMap == MAP_FAILED)
	{
		Close();
		SetError("Unable to map view for the memory map file.");
		return false;
	}
	else
	{
		m_lpData = (char *)(m_lpFileMap) + cellf_page_offset;
	}
	if (m_File != NULL) 
	{
		fclose(m_File);
		m_File = NULL;
	}
#endif

	// Read the Mean data
	m_pMeanIntensities = (unsigned short*)m_lpData;

	retVal = true;
	return retVal;
}

//////////////////////////////////////////////////////////////////////
  
void CCELFileData::EnsureNotMmapped() 
{
	// skip if not mmapped...
	if (m_lpFileMap==NULL)
	{
		return;
	}

	// it is mmapped, so fix it.
	if (m_FileFormat == TRANSCRIPTOME_BCEL) // the (short,short,char) format
	{ 
		assert(m_pTransciptomeEntries != NULL);

		// duplicate the data
		size_t bytecnt = GetCols() * GetRows() * sizeof(CELFileTranscriptomeEntryType);
		CELFileTranscriptomeEntryType *tmpptr = (CELFileTranscriptomeEntryType *) malloc(bytecnt);
		memcpy(tmpptr, m_pTransciptomeEntries, bytecnt);

		// discard the old map
		Munmap();

		// put it in place
		m_pTransciptomeEntries = tmpptr;
	} 
	else if (m_FileFormat == XDA_BCEL) // the (float,float,short) format
	{ 
		assert(m_pEntries != NULL);

		// duplicate the data
		size_t bytecnt = GetCols() * GetRows() * sizeof(CELFileEntryType);
		CELFileEntryType *tmpptr = (CELFileEntryType *) malloc(bytecnt);
		memcpy(tmpptr, m_pEntries, bytecnt);

		// discard the old map
		Munmap();

		// put it in place
		m_pEntries = tmpptr;
	}
	else if (m_FileFormat == COMPACT_BCEL) // the (short) format
	{ 
		assert(m_pMeanIntensities != NULL);

		// duplicate the data
		size_t bytecnt = GetCols() * GetRows() * sizeof(unsigned short);
		unsigned short *tmpptr = (unsigned short *) malloc(bytecnt);
		memcpy(tmpptr, m_pMeanIntensities, bytecnt);

		// discard the old map
		Munmap();

		// put it in place
		m_pMeanIntensities = tmpptr;
	}
}

//////////////////////////////////////////////////////////////////////

bool CCELFileData::ReadTextCel(bool bReadHeaderOnly)
{
	bool retVal = false;

	// Open the file.
	std::ifstream instr;
	instr.open(m_FullPathName.c_str(), std::ios::in);

	// Check if open
	if (!instr)
	{
		SetError("Unable to open the file.");
		return false;
	}

	Clear();

	const int MAXLINELENGTH = 4096;
	char pszHeader[MAXLINELENGTH];
	char tempStr[MAXLINELENGTH];

	// Extract a line of header
	instr.getline(pszHeader, MAXLINELENGTH);

	//Determine the version number
	if (strncmp(pszHeader,"[CEL]",5)==0)
		m_HeaderData.SetVersion(3);
	else if (strncmp(pszHeader,"COLS/ROWS=",10)==0)
		m_HeaderData.SetVersion(2);
	else
	{
		SetError("Unrecognized CEL file format.");
		return false;
	}

	//read and store the header
	if (m_HeaderData.GetVersion() == 2)
	{
		m_HeaderData.SetHeader(pszHeader);
		instr.getline(pszHeader, MAXLINELENGTH);
		m_HeaderData.AppendHeader(pszHeader);
		instr.getline(pszHeader, MAXLINELENGTH);

		// Store rows and columns
		int iCols = 0;
		int iRows = 0;
		sscanf(m_HeaderData.GetHeader().c_str(), "COLS/ROWS=%d %d", &iCols, &iRows);
		m_HeaderData.SetCols(iCols);
		m_HeaderData.SetRows(iRows);
	}
	else
	{
		bool moreSpace=true;
		//Read past [HEADER] to first line of header data
		while (moreSpace)
		{
			if (!instr.getline(pszHeader, MAXLINELENGTH))
				return false;
			if (strncmp(pszHeader,"[HEADER]",8)==0)
				moreSpace=false;
		}
		int iCols = 0;
		int iRows = 0;
		instr.getline(pszHeader, MAXLINELENGTH);
		sscanf(pszHeader,"Cols=%d",&iCols);
		m_HeaderData.SetCols(iCols);
		m_HeaderData.SetHeader(pszHeader);
		instr.getline(pszHeader, MAXLINELENGTH);
		sscanf(pszHeader,"Rows=%d",&iRows);
		m_HeaderData.SetRows(iRows);
		m_HeaderData.AppendHeader(pszHeader);

		//Now read the rest of the header
		bool moreHeader=true;
		while (moreHeader)
		{
			if (!instr.getline(pszHeader, MAXLINELENGTH))
				return false;
			if (strncmp(pszHeader,"DatHeader=",10)==0) //Last line of the header
			{
                if (pszHeader[strlen(pszHeader) - 1] == '\r') pszHeader[strlen(pszHeader) - 1] = '\0';
				m_HeaderData.SetDatHeader(pszHeader + 10);
			}

			if (strncmp(pszHeader,"Algorithm=",10)==0) //Last line of the header
			{
				sscanf(pszHeader,"Algorithm=%s",tempStr);
				m_HeaderData.SetAlg(tempStr);
			}

			if (strncmp(pszHeader,"AlgorithmParameters=",20)==0) //Last line of the header
			{
				sscanf(pszHeader,"AlgorithmParameters=%s",tempStr);
				m_HeaderData.SetParams(tempStr);
				moreHeader=false;
			}
			m_HeaderData.AppendHeader(pszHeader);
		}
	}

	// Set the chip type
	m_HeaderData.ParseChipType();

	// Parse algorithm parameters into map
	m_HeaderData.ParseAlgorithmParameters();

	// Set grid coordinates
	m_HeaderData.ParseCorners();

	char paramString[1024];
	GridCoordinatesType grid = m_HeaderData.GetCorners();
	sprintf(paramString, "Cols=%d\nRows=%d\nTotalX=%d\nTotalY=%d\nOffsetX=%d\nOffsetY=%d\n"
				"GridCornerUL=%ld %ld\nGridCornerUR=%ld %ld\nGridCornerLR=%ld %ld\nGridCornerLL=%ld %ld\n"
				"Axis-invertX=%d\nAxisInvertY=%d\nswapXY=%d\nDatHeader=%s"
				"\nAlgorithm=%s\nAlgorithmParameters=%s\n",
				m_HeaderData.GetCols(),m_HeaderData.GetRows(),
				m_HeaderData.GetCols(),m_HeaderData.GetRows(),
				0,0,
				grid.upperleft.x, grid.upperleft.y,
				grid.upperright.x, grid.upperright.y,
				grid.lowerright.x, grid.lowerright.y,
				grid.lowerleft.x, grid.lowerleft.y,
				0,0,0,
				m_HeaderData.GetDatHeader().c_str(),
				m_HeaderData.GetAlg().c_str(),
				m_HeaderData.GetAlgorithmParameters().c_str());

	m_HeaderData.SetHeader(paramString);

	// Don't continue if just reading the header.
	if (bReadHeaderOnly)
		return true;

	// Create memory for Mean data.
	m_HeaderData.SetCells(m_HeaderData.GetRows() * m_HeaderData.GetCols());
	m_pEntries = new CELFileEntryType[m_HeaderData.GetCells()];

  // tmps
  int t_x,t_y,t_pixels;
  float t_mean,t_stdv;

	// Read v2 CEL files
	if (m_HeaderData.GetVersion() == 2)
	{
		// Write the Mean data
		const char *strCellEntryFormat="%d %d %f %f %d";
		for (int iCell=0; iCell < m_HeaderData.GetCells(); iCell++)
		{
			instr.getline(pszHeader, MAXLINELENGTH);
			CELFileEntryType *pEntry = m_pEntries + iCell;
      // FIXME: Check return value of sscanf
			sscanf(pszHeader, strCellEntryFormat,          
             &t_x,
             &t_y,
             &t_mean,
             &t_stdv,
             &t_pixels);
      SetMean(t_x,t_y,t_mean);
      SetStdev(t_x,t_y,t_stdv);
      SetPixels(t_x,t_y,t_pixels);
		}
		retVal = true;
	}
	else
	{
		//Advance to the beginning of the Mean data
		bool readMore=true;
		while(readMore)
		{
			if (!instr.getline(pszHeader, MAXLINELENGTH))
				return false;
			if (strncmp(pszHeader,"[INTENSITY]",11)==0)
				readMore=false;
		}
  		instr.getline(pszHeader, MAXLINELENGTH);//Data starts at 2 lines past [Mean]
  		instr.getline(pszHeader, MAXLINELENGTH);//Data starts at 2 lines past [Mean]

		//Read the Mean data
		int iCell=0;
		readMore=true;
		const char *strCellEntryFormat="%d\t%d\t%f\t%f\t%d";

		//int x,y,pix; // unused.

		while (readMore)
		{
			if (!instr.getline(pszHeader, MAXLINELENGTH)) //end of file
				readMore=false;
			else if (strlen(pszHeader) < MIN_CELLSTR )// blank line at end of data
				readMore=false;
			else
			{
				CELFileEntryType *pEntry = m_pEntries + iCell;
				sscanf(pszHeader, strCellEntryFormat,          
             &t_x,
             &t_y,
             &t_mean,
             &t_stdv,
             &t_pixels);
        SetMean(t_x,t_y,t_mean);
        SetStdev(t_x,t_y,t_stdv);
        SetPixels(t_x,t_y,t_pixels);
				++iCell;
			}
		}

		//Advance to the Masked data
		readMore=true;
		while(readMore)
		{
			if (!instr.getline(pszHeader, MAXLINELENGTH)) //end of file
				return false;
			if (strncmp(pszHeader,"[MASKS]",7)==0)
				readMore=false;
		}
		//Read number of masked cells
		instr.getline(pszHeader, MAXLINELENGTH);
		int nMasked=0;
		sscanf(pszHeader, "NumberCells=%d", &nMasked);

		m_HeaderData.SetMasked(nMasked);
		instr.getline(pszHeader, MAXLINELENGTH);//skip over the header

		//Read the masked data
		if (m_bReadMaskedCells)
		{
			readMore=true;
			while (readMore)
			{
				if (!instr.getline(pszHeader, MAXLINELENGTH)) //end of file
					readMore=false;
				else if (strlen(pszHeader) < MIN_CELLSTR )// blank line at end of data
					readMore=false;
				else
				{
					int x, y, iCellEntry;
					sscanf(pszHeader, "%d\t%d", &x, &y);
					iCellEntry = y * m_HeaderData.GetCols() + x;
					m_MaskedCells.insert(std::make_pair(iCellEntry, true));
				}
			}
		}

		//Advance to the outlier data
		readMore=true;
		while(readMore)
		{
			if (!instr.getline(pszHeader, MAXLINELENGTH)) //end of file
				return false;
			if (strncmp(pszHeader,"[OUTLIERS]",10)==0)
				readMore=false;
		}
		//Read number of outlier cells
		instr.getline(pszHeader, MAXLINELENGTH);
		int nOutliers=0;
		sscanf(pszHeader, "NumberCells=%d", &nOutliers);
		m_HeaderData.SetOutliers(nOutliers);
		instr.getline(pszHeader, MAXLINELENGTH);//skip over the header

		//Read the outlier data
		if (m_bReadOutliers)
		{
			readMore=true;
			while (readMore)
			{
				if (!instr.getline(pszHeader, MAXLINELENGTH)) //end of file
					readMore=false;
				else if (strlen(pszHeader) < MIN_CELLSTR )// blank line at end of data
					readMore=false;
				else
				{
					int x, y, iCellEntry;
					sscanf(pszHeader, "%d\t%d", &x, &y);
					iCellEntry = y * m_HeaderData.GetCols() + x;
					m_Outliers.insert(std::make_pair(iCellEntry, true));
				}
			}
		}
		retVal = true;
	}
	instr.close();
	return retVal;
}

//////////////////////////////////////////////////////////////////////

void CCELFileData::Clear()
{
	m_HeaderData.Clear();
	if (m_lpData == NULL)
	{
		delete [] m_pEntries;
		delete [] m_pTransciptomeEntries;
		delete [] m_pMeanIntensities;
	}
	m_pEntries = NULL;
	m_pTransciptomeEntries = NULL;
	m_pMeanIntensities = NULL;
	m_MaskedCells.clear();
	m_Outliers.clear();
}

////////////////////////////////////////////////////////////////////////

void CCELFileData::Munmap() 
{
	// This is a nop on non-mmapped files.
	if (m_lpFileMap == NULL)
	{
		return;
	}

	// zero out the pointers.
	m_pTransciptomeEntries = NULL;
	m_pEntries = NULL;
	m_pMeanIntensities = NULL;

	// free the map
#ifdef WIN32
	if (m_lpFileMap != NULL)
	{
		UnmapViewOfFile(m_lpFileMap);
		CloseHandle(m_hFileMap);
		m_hFileMap = NULL;
		CloseHandle(m_hFile);
		m_hFile = INVALID_HANDLE_VALUE;
	}
#else
	if ((m_lpFileMap != MAP_FAILED) && (m_lpFileMap != NULL))
	{
		munmap(m_lpFileMap, m_MapLen);
		m_lpFileMap = MAP_FAILED;
		m_MapLen = 0;
		// should be closed already.
		if (m_File != NULL) 
		{
			fclose(m_File);
			m_File = NULL;
		}
	}
#endif
}

void CCELFileData::Close()
{
	Clear();
	Munmap();
	m_lpData = NULL;
}

//////////////////////////////////////////////////////////////////////

bool CCELFileData::ReadHeader()
{
	// Read the header, close if failed.
	if (Open(true) == false)
	{
		Close();
		return false;
	}
	return true;
}

//////////////////////////////////////////////////////////////////////

bool CCELFileData::Read()
{
	// Open the file
	if (Open() == false)
	{
		Close();
		return false;
	}
	return true;
}

//////////////////////////////////////////////////////////////////////

bool CCELFileData::ReadEx(const char *filename, int nState)
{
	// Open the file
	SetFileName(filename);

	// Determine if outliers should be read
	if ((nState & (CEL_ALL | CEL_OUTLIER)))
		this->m_bReadOutliers = true;
	else
		this->m_bReadOutliers = false;

	// Determine if masks should be read
	if ((nState & (CEL_ALL | CEL_MASK)))
		this->m_bReadMaskedCells = true;
	else
		this->m_bReadMaskedCells = false;

	if (Open() == false)
	{
		Close();
		return false;
	}
	return true;
}

//////////////////////////////////////////////////////////////////////
std::string CCELFileData::GetHeaderKey(const char* key)
{
	assert(key != NULL);

	std::string strKey = key;
	std::transform(strKey.begin(), strKey.end(), strKey.begin(), toupper);

	if (strKey == "HEADER")
		return GetHeader();
	else if (strKey == "VERSION")
	{
		char* sValue = new char[50];
		sprintf(sValue, "%d", GetVersion());
		return sValue;
	}
	else if (strKey == "COLS")
	{
		char* sValue = new char[50];
		sprintf(sValue, "%d", GetCols());
		return sValue;
	}
	else if (strKey == "ROWS")
	{
		char* sValue = new char[50];
		sprintf(sValue, "%d", GetRows());
		return sValue;
	}
	else if (strKey == "TOTALX")
	{
		char* sValue = new char[50];
		sprintf(sValue, "%d", GetCols());
		return sValue;
	}
	else if (strKey == "TOTALY")
	{
		char* sValue = new char[50];
		sprintf(sValue, "%d", GetRows());
		return sValue;
	}
	else if (strKey == "GRIDCORNERUL")
	{
		char* sValue = new char[50];
		sprintf(sValue, "(%ld, %ld)", GetCorners().upperleft.x, GetCorners().upperleft.y);
		return sValue;
	}
	else if (strKey == "GRIDCORNERUR")
	{
		char* sValue = new char[50];
		sprintf(sValue, "(%ld, %ld)", GetCorners().upperright.x, GetCorners().upperright.y);
		return sValue;
	}
	else if (strKey == "GRIDCORNERLL")
	{
		char* sValue = new char[50];
		sprintf(sValue, "(%ld, %ld)", GetCorners().lowerleft.x, GetCorners().lowerleft.y);
		return sValue;
	}
	else if (strKey == "GRIDCORNERLR")
	{
		char* sValue = new char[50];
		sprintf(sValue, "(%ld, %ld)", GetCorners().lowerright.x, GetCorners().lowerright.y);
		return sValue;
	}
	else if (strKey == "OFFSETX")
		return "0";
	else if (strKey == "OFFSETY")
		return "0";
	else if (strKey == "AXIS-INVERTX")
		return "0";
	else if (strKey == "AXISINVERTY")
		return "0";
	else if (strKey == "SWAPXY")
		return "0";
	else if (strKey == "DATHEADER")
		return GetDatHeader();
	else if (strKey == "ALGORITHM")
		return GetAlg();
	else if (strKey == "ALGORITHMPARAMETERS")
		return GetParams();
	else if (strKey == "NUMBERCELLS")
	{
		char* sValue = new char[50];
		sprintf(sValue, "%d", GetNumCells());
		return sValue;
	}
	else if (strKey == "NUMBERMASKEDCELLS")
	{
		char* sValue = new char[50];
		sprintf(sValue, "%ld", GetNumMasked());
		return sValue;
	}
	else if (strKey == "NUMBEROUTLIERCELLS")
	{
		char* sValue = new char[50];
		sprintf(sValue, "%ld", GetNumOutliers());
		return sValue;
	}
	else
	{
		return NULL;
	}
}

//////////////////////////////////////////////////////////////////////

float CCELFileData::GetMean(int x, int y)
{
	assert((x >= 0) && (x <= m_HeaderData.GetCols()));
	assert((y >= 0) && (y <= m_HeaderData.GetRows()));

	float fMean = 0;
	int index = y * m_HeaderData.GetCols() + x;

	if (m_FileFormat == TEXT_CEL) {
		//fMean = m_pEntries[index].Mean;
    fMean=MmGetFloat_I(&m_pEntries[index].Mean);
  }
	else if (m_FileFormat == XDA_BCEL) {
		//fMean = GetFloat(&(m_pEntries[index].Mean), m_FileFormat);
    fMean=MmGetFloat_I(&(m_pEntries[index].Mean));
  }
	else if (m_FileFormat == TRANSCRIPTOME_BCEL) {
		//fMean = (float) (GetUShort(&(m_pTransciptomeEntries[index].Mean), m_FileFormat));
    fMean=MmGetUInt16_N(&(m_pTransciptomeEntries[index].Mean));
  }
	else if (m_FileFormat == COMPACT_BCEL) {
		//fMean = (float) (GetUShort(m_pMeanIntensities + index, m_FileFormat));
    fMean=MmGetUInt16_I(&m_pMeanIntensities[index]);
  }
  else {
    assert(0);
  }

	return fMean;
}

void CCELFileData::SetMean(int x, int y, float mean)
{
	assert((x >= 0) && (x <= m_HeaderData.GetCols()));
	assert((y >= 0) && (y <= m_HeaderData.GetRows()));

	int index = (m_HeaderData.GetCols() * y) + x;

  if (m_FileFormat == TEXT_CEL) {
    MmSetFloat_I(&m_pEntries[index].Mean,mean);
  }
  else if (m_FileFormat == XDA_BCEL) {
		//m_pEntries[index].Mean = GetFloat(&mean, m_FileFormat);
    MmSetFloat_I(&m_pEntries[index].Mean,mean);
  }
  else if (m_FileFormat == TRANSCRIPTOME_BCEL)	{
		unsigned short s = RoundNumber(mean);
    MmSetUInt16_N(&m_pTransciptomeEntries[index].Mean,s);
  }
	else if (m_FileFormat == COMPACT_BCEL) {
		unsigned short s = RoundNumber(mean);
		//m_pMeanIntensities[index] = GetUShort(&s, m_FileFormat);
    MmSetUInt16_I(&m_pMeanIntensities[index],s);
	}
  else {
    assert(0);
  }
}



float CCELFileData::GetStdev(int x, int y)
{
	assert((x >= 0) && (x <= m_HeaderData.GetCols()));
	assert((y >= 0) && (y <= m_HeaderData.GetRows()));

	float fStdev = 0;
	int index = y * m_HeaderData.GetCols() + x;
	if (m_FileFormat == TEXT_CEL)
		//fStdev = m_pEntries[index].Stdv;
		fStdev=MmGetFloat_I(&m_pEntries[index].Stdv);
	else if (m_FileFormat == XDA_BCEL)
		//fStdev = GetFloat(&(m_pEntries[index].Stdv), m_FileFormat);
		fStdev=MmGetFloat_I(&(m_pEntries[index].Stdv));
	else if (m_FileFormat == TRANSCRIPTOME_BCEL)
		//fStdev = (float) (GetUShort(&(m_pTransciptomeEntries[index].Stdv), m_FileFormat));
		fStdev=MmGetUInt16_N(&(m_pTransciptomeEntries[index].Stdv));
	else if (m_FileFormat == COMPACT_BCEL)
		fStdev=0;
  else 
    assert(0);

	return fStdev;
}

short CCELFileData::GetPixels(int x, int y)
{
	assert((x >= 0) && (x <= m_HeaderData.GetCols()));
	assert((y >= 0) && (y <= m_HeaderData.GetRows()));

	short sPixels = 0;
	int index = y * m_HeaderData.GetCols() + x;
	if (m_FileFormat == TEXT_CEL)
		//sPixels = m_pEntries[index].Pixels;
		sPixels=MmGetInt16_I(&m_pEntries[index].Pixels);
	else if (m_FileFormat == XDA_BCEL)
		//sPixels = GetShort(&(m_pEntries[index].Pixels), m_FileFormat);
		sPixels=MmGetInt16_I(&m_pEntries[index].Pixels);
	else if (m_FileFormat == TRANSCRIPTOME_BCEL)
		//sPixels = (short) m_pTransciptomeEntries[index].Pixels;
		sPixels=MmGetUInt8(&m_pTransciptomeEntries[index].Pixels);
	else if (m_FileFormat == COMPACT_BCEL)
		sPixels=0;
  else 
    assert(0);

	return sPixels;
}

bool CCELFileData::IsMasked(int x, int y)
{
	assert((x >= 0) && (x <= m_HeaderData.GetCols()));
	assert((y >= 0) && (y <= m_HeaderData.GetRows()));

	int index = y * m_HeaderData.GetCols() + x;
	bool bReturn = false;
	std::map<int, bool>::iterator pos = m_MaskedCells.find(index);
	if (pos != m_MaskedCells.end())
		bReturn = true;
	return bReturn;
}

bool CCELFileData::IsOutlier(int x, int y)
{
	assert((x >= 0) && (x <= m_HeaderData.GetCols()));
	assert((y >= 0) && (y <= m_HeaderData.GetRows()));

	int index = y * m_HeaderData.GetCols() + x;
	bool bReturn = false;
	std::map<int, bool>::iterator pos = m_Outliers.find(index);
	if (pos != m_Outliers.end())
		bReturn = true;
	return bReturn;
}

//////////////////////////////////////////////////////////////////////

bool CCELFileData::IsXDACompatibleFile()
{
	// Open the file.
	std::ifstream instr(m_FullPathName.c_str(), std::ios::in | std::ios::binary);
	if (!instr)
		return 0;

	// Read the magic number from the file.
	uint32_t magic=0;
	ReadUInt32_I(instr, magic);
	bool bXDAFile = ((magic == CELL_FILE_MAGIC_NUMBER) ? true : false);
	instr.close();

	return bXDAFile;
}

bool CCELFileData::IsTranscriptomeBcelFile()
{
	// Open the file.
	std::ifstream instr(m_FullPathName.c_str(), std::ios::in | std::ios::binary);
	if (!instr)
		return 0;

	// Read the header marker from the file.
	char szMarker[BCEL_HEADER_LEN];
	READ_FIXED_STRING(instr, szMarker, BCEL_HEADER_LEN);
	bool bTranscriptomeFile = false;
	if (strncmp(szMarker, BCEL_HEADER_BYTES, BCEL_HEADER_LEN) == 0)
		bTranscriptomeFile = true;
	instr.close();

	return bTranscriptomeFile;
}

bool CCELFileData::IsCompactCelFile()
{
	// Open the file.
	std::ifstream instr(m_FullPathName.c_str(), std::ios::in | std::ios::binary);
	if (!instr)
		return 0;

	// Read the header marker from the file.
	char szMarker[CCEL_HEADER_LEN];
	READ_FIXED_STRING(instr, szMarker, CCEL_HEADER_LEN);
	bool bCompactCelFile = false;
	if (strncmp(szMarker, CCEL_HEADER_BYTES, CCEL_HEADER_LEN) == 0)
		bCompactCelFile = true;
	instr.close();

	return bCompactCelFile;
}

void CCELFileData::DetermineFileFormat()
{
	if (IsXDACompatibleFile())
		m_FileFormat = XDA_BCEL;
	else if (IsTranscriptomeBcelFile())
		m_FileFormat = TRANSCRIPTOME_BCEL;
	else if (IsCompactCelFile())
		m_FileFormat = COMPACT_BCEL;
	else
		m_FileFormat = TEXT_CEL;
}


unsigned short CCELFileData::RoundNumber(float f)
{
	assert(f >= 0);
	unsigned short us = (unsigned short) f;
	if((f - (float) us) >= 0.5)
		us++;
	return us;
}

//////////////////////////////////////////////////////////////////////

int CCELFileData::GetX(int index)
{
	return index % m_HeaderData.GetCols();
}

int CCELFileData::GetY(int index)
{
	return index / m_HeaderData.GetCols();
}

//////////////////////////////////////////////////////////////////////

CCELFileData::CCELFileData()
{
	m_strError = "";
	m_FullPathName = "";
	m_LibPath = "";
	m_DataPath = "";
	m_FileName = "";
	m_pEntries = NULL;
	m_pTransciptomeEntries = NULL;
	m_pMeanIntensities = NULL;
	m_FileFormat = XDA_BCEL;
	m_lpFileMap = NULL;
	m_lpData = NULL;
	m_bReadMaskedCells = true;
	m_bReadOutliers = true;
	m_nReadState = CEL_ALL;
}

//////////////////////////////////////////////////////////////////////

CCELFileData::~CCELFileData()
{
	Clear();
}

//////////////////////////////////////////////////////////////////////

void CCELFileData::SetError(const char* str)
{
	assert(str != NULL);
	int iLen = (int) strlen(str);
	assert(iLen > 0);
	m_strError = str;
}

//////////////////////////////////////////////////////////////////////

void CCELFileData::SetAlgorithmName(const char *str)
{
	m_HeaderData.SetAlg(str);
}

//////////////////////////////////////////////////////////////////////

void CCELFileData::AddAlgorithmParameter(const char *tag, const char *value)
{
	m_HeaderData.AddAlgorithmParameter(tag, value);
}

//////////////////////////////////////////////////////////////////////

void CCELFileData::SetFileFormat(int i) 
{ 
	assert((i >= TEXT_CEL) && (i <= COMPACT_BCEL)); 
	m_FileFormat = i; 
}

//////////////////////////////////////////////////////////////////////

void CCELFileData::SetDimensions(int rows, int cols)
{
	m_HeaderData.SetRows(rows);
	m_HeaderData.SetCols(cols);
	m_HeaderData.SetCells(rows * cols);
	GridCoordinatesType grid;
	grid.upperleft.x = 1;
	grid.upperleft.y = 1;
	grid.upperright.x = cols;
	grid.upperright.y = 1;
	grid.lowerleft.x = 1;
	grid.lowerleft.y = rows;
	grid.lowerright.x = cols;
	grid.lowerright.y = rows;
	m_HeaderData.SetGrid(grid);
	delete [] m_pEntries;
	delete [] m_pTransciptomeEntries;
	if (m_FileFormat == TRANSCRIPTOME_BCEL)
		m_pTransciptomeEntries = new CELFileTranscriptomeEntryType[rows*cols];
	else if ((m_FileFormat == XDA_BCEL) || (m_FileFormat == TEXT_CEL))
		m_pEntries = new CELFileEntryType[rows*cols];
	else if (m_FileFormat == COMPACT_BCEL)
		m_pMeanIntensities = new unsigned short[rows*cols];
}

//////////////////////////////////////////////////////////////////////

void CCELFileData::SetChipType(const char *str)
{
	assert(str != NULL);
	m_HeaderData.SetChipType(str);
	
	std::string strHeader = m_HeaderData.GetHeader();

	// Set the chip type in the header string.
	int endIndex = (int) strHeader.find(".1sq");
	int startIndex = (int) strHeader.rfind(" ", endIndex);
	int length = (int) strHeader.length();

	std::string start = strHeader.substr(0, startIndex + 1);
	std::string end = strHeader.substr(endIndex, length - endIndex);
	std::string newHeader =	start + m_HeaderData.GetChipType() + end;
	m_HeaderData.SetHeader(newHeader.c_str());
}

//////////////////////////////////////////////////////////////////////

void CCELFileData::SetMargin(int margin)
{
	m_HeaderData.SetMargin(margin);
}

//////////////////////////////////////////////////////////////////////

void CCELFileData::SetStdev(int x, int y, float stdev)
{
	assert((x >= 0) && (x <= m_HeaderData.GetCols()));
	assert((y >= 0) && (y <= m_HeaderData.GetRows()));

	int index = (m_HeaderData.GetCols() * y) + x;
	if (m_FileFormat == TRANSCRIPTOME_BCEL)
	{
		unsigned short s = RoundNumber(stdev);
		//m_pTransciptomeEntries[index].Stdv = GetUShort(&s, m_FileFormat);
		MmSetUInt16_N(&m_pTransciptomeEntries[index].Stdv,s);
	}
	else if ((m_FileFormat == TEXT_CEL) || (m_FileFormat == XDA_BCEL))
		//m_pEntries[index].Stdv = GetFloat(&stdev, m_FileFormat);
		MmSetFloat_I(&m_pEntries[index].Stdv,stdev);
	else if (m_FileFormat != COMPACT_BCEL) 
	    assert(0);
}

void CCELFileData::SetPixels(int x, int y, short pixels)
{
	assert((x >= 0) && (x <= m_HeaderData.GetCols()));
	assert((y >= 0) && (y <= m_HeaderData.GetRows()));

	int index = (m_HeaderData.GetCols() * y) + x;
	if (m_FileFormat == TRANSCRIPTOME_BCEL)
		//m_pTransciptomeEntries[index].Pixels = (unsigned char) pixels;
    MmSetUInt8(&m_pTransciptomeEntries[index].Pixels,(uint8_t)pixels);
	else if ((m_FileFormat == TEXT_CEL) || (m_FileFormat == XDA_BCEL))
		//m_pEntries[index].Pixels = GetShort(&pixels, m_FileFormat);
		MmSetUInt16_I((uint16_t*)&m_pEntries[index].Pixels,(uint16_t)pixels);
	else if (m_FileFormat != COMPACT_BCEL) 
	    assert(0);
}

void CCELFileData::SetMasked(int x, int y, bool masked)
{
	assert((x >= 0) && (x <= m_HeaderData.GetCols()));
	assert((y >= 0) && (y <= m_HeaderData.GetRows()));

	int index = (m_HeaderData.GetCols() * y) + x;
	if (masked)
	{
		m_MaskedCells.insert(std::make_pair(index, true));
		m_HeaderData.IncrementMasked();
	}
	else
	{
		std::map<int, bool>::iterator pos = m_MaskedCells.find(index);
		if (pos != m_MaskedCells.end())
		{
			m_MaskedCells.erase(pos);
			m_HeaderData.DecrementMasked();
		}
	}
}

void CCELFileData::SetOutlier(int x, int y, bool outlier)
{
	assert((x >= 0) && (x <= m_HeaderData.GetCols()));
	assert((y >= 0) && (y <= m_HeaderData.GetRows()));

	int index = (m_HeaderData.GetCols() * y) + x;
	if (outlier)
	{
		m_Outliers.insert(std::make_pair(index, true));
		m_HeaderData.IncrementOutliers();
	}
	else
	{
		std::map<int, bool>::iterator pos = m_Outliers.find(index);
		if (pos != m_Outliers.end())
		{
			m_Outliers.erase(pos);
			m_HeaderData.DecrementOutliers();
		}
	}
}

//////////////////////////////////////////////////////////////////////

bool CCELFileData::WriteTextCel()
{
	if (m_FullPathName.length() == 0)
	{
		SetError("No file name is set for file creation.");
		return false;
	}

	std::ofstream newCelFile;
	newCelFile.open(m_FullPathName.c_str(), std::ios::out);
	if (!newCelFile)
	{
		SetError("Unable to open the file.");
		return false;
	}

	m_HeaderData.SetDatHeader();

	const int MAXLINELENGTH = 4096;
	char szBuffer[MAXLINELENGTH] = "\0";
	int length = 0;

	// Write CEL 
	length = sprintf(szBuffer, "[CEL]%s", LINE_SEPARATOR);
	newCelFile.write(szBuffer, length);
	length = sprintf(szBuffer, "Version=3%s%s", LINE_SEPARATOR, LINE_SEPARATOR);
	newCelFile.write(szBuffer, length);

	// Write the header.
	GridCoordinatesType grid = m_HeaderData.GetCorners();
	length = sprintf(szBuffer, "[HEADER]%s", LINE_SEPARATOR);
	newCelFile.write(szBuffer, length);
	length = sprintf(szBuffer, "Cols=%d%sRows=%d%s", m_HeaderData.GetCols(), LINE_SEPARATOR, m_HeaderData.GetRows(), LINE_SEPARATOR);
	newCelFile.write(szBuffer, length);
	length = sprintf(szBuffer, "TotalX=%d%sTotalY=%d%s", m_HeaderData.GetCols(), LINE_SEPARATOR, m_HeaderData.GetRows(), LINE_SEPARATOR);
	newCelFile.write(szBuffer, length);
	length = sprintf(szBuffer, "OffsetX=0%sOffsetY=0%s", LINE_SEPARATOR, LINE_SEPARATOR);
	newCelFile.write(szBuffer, length);
	length = sprintf(szBuffer, "GridCornerUL=%ld %ld%s", grid.upperleft.x, grid.upperleft.y, LINE_SEPARATOR);
	newCelFile.write(szBuffer, length);
	length = sprintf(szBuffer, "GridCornerUR=%ld %ld%s", grid.upperright.x, grid.upperright.y, LINE_SEPARATOR);
	newCelFile.write(szBuffer, length);
	length = sprintf(szBuffer, "GridCornerLR=%ld %ld%s", grid.lowerright.x, grid.lowerright.y, LINE_SEPARATOR);
	newCelFile.write(szBuffer, length);
	length = sprintf(szBuffer, "GridCornerLL=%ld %ld%s", grid.lowerleft.x, grid.lowerleft.y, LINE_SEPARATOR);
	newCelFile.write(szBuffer, length);
	length = sprintf(szBuffer, "Axis-invertX=0%sAxisInvertY=0%sswapXY=0%s", LINE_SEPARATOR, LINE_SEPARATOR, LINE_SEPARATOR);
	newCelFile.write(szBuffer, length);
	length = sprintf(szBuffer, "DatHeader=%s%s", m_HeaderData.GetDatHeader().c_str(), LINE_SEPARATOR);
	newCelFile.write(szBuffer, length);
	length = sprintf(szBuffer, "Algorithm=%s%s", m_HeaderData.GetAlg().c_str(), LINE_SEPARATOR);
	newCelFile.write(szBuffer, length);
	length = sprintf(szBuffer, "AlgorithmParameters=%s%s%s", m_HeaderData.GetAlgorithmParameters().c_str(), LINE_SEPARATOR, LINE_SEPARATOR);
	newCelFile.write(szBuffer, length);

	// Write the Mean header
	length = sprintf(szBuffer, "[INTENSITY]%s", LINE_SEPARATOR);
	newCelFile.write(szBuffer, length);
	length = sprintf(szBuffer, "NumberCells=%d%s", m_HeaderData.GetCells(), LINE_SEPARATOR);
	newCelFile.write(szBuffer, length);
	length = sprintf(szBuffer, "CellHeader=X\tY\tMEAN\tSTDV\tNPIXELS%s", LINE_SEPARATOR);
	newCelFile.write(szBuffer, length);

  //
	for (int iCell=0; iCell < m_HeaderData.GetCells(); iCell++)
	{
    int   t_x=GetX(iCell);
    int   t_y=GetY(iCell);
    float t_mean=GetMean(t_x,t_y);
    float t_stdv=GetStdev(t_x,t_y);
    int   t_pixel=GetPixels(t_x,t_y);

    sprintf(szBuffer, "%3d\t%3d\t%.1f\t%.1f\t%3d",t_x,t_y,t_mean,t_stdv,t_pixel);
    newCelFile<<szBuffer<<LINE_SEPARATOR;
	}
  newCelFile<<LINE_SEPARATOR;
  

	// Write the mask data
	length = sprintf(szBuffer, "[MASKS]%s", LINE_SEPARATOR);
	newCelFile.write(szBuffer, length);
	length = sprintf(szBuffer, "NumberCells=%ld%s", m_HeaderData.GetMasked(), LINE_SEPARATOR);
	newCelFile.write(szBuffer, length);
	length = sprintf(szBuffer, "CellHeader=X\tY%s", LINE_SEPARATOR);
	newCelFile.write(szBuffer, length);
	for(std::map<int, bool>::iterator pos = m_MaskedCells.begin(); pos != m_MaskedCells.end(); pos++)
	{
		length = sprintf(szBuffer, "%d\t%d%s", GetX(pos->first), GetY(pos->first), LINE_SEPARATOR);
		newCelFile.write(szBuffer, length);
	}
	length = sprintf(szBuffer, "%s", LINE_SEPARATOR);
	newCelFile.write(szBuffer, length);

	// Write the outlier data
	length = sprintf(szBuffer, "[OUTLIERS]%s", LINE_SEPARATOR);
	newCelFile.write(szBuffer, length);
	length = sprintf(szBuffer, "NumberCells=%ld%s", m_HeaderData.GetOutliers(), LINE_SEPARATOR);
	newCelFile.write(szBuffer, length);
	length = sprintf(szBuffer, "CellHeader=X\tY%s", LINE_SEPARATOR);
	newCelFile.write(szBuffer, length);
	for(std::map<int, bool>::iterator pos = m_Outliers.begin(); pos != m_Outliers.end(); pos++)
	{
		length = sprintf(szBuffer, "%d\t%d%s", GetX(pos->first), GetY(pos->first), LINE_SEPARATOR);
		newCelFile.write(szBuffer, length);
	}

	// Close the file and check the status.
	newCelFile.close();

	return !newCelFile.fail();
}

//////////////////////////////////////////////////////////////////////

bool CCELFileData::WriteXDABCel()
{
	if (m_FullPathName.length() == 0)
	{
		SetError("No file name is set for file creation.");
		return false;
	}

	std::ofstream newCelFile;
	newCelFile.open(m_FullPathName.c_str(), std::ios::out | std::ios::binary);
	if (!newCelFile)
	{
		SetError("Unable to open the file.");
		return false;
	}

	m_HeaderData.SetDatHeader();

	// Write the header.
	WriteUInt32_I(newCelFile, CELL_FILE_MAGIC_NUMBER);
	WriteUInt32_I(newCelFile, CELL_FILE_VERSION_NUMBER);

	// Write the dimensions of the array
	WriteUInt32_I(newCelFile, m_HeaderData.GetRows());
	WriteUInt32_I(newCelFile, m_HeaderData.GetCols());
	WriteUInt32_I(newCelFile, m_HeaderData.GetCells());

	// Write the other members.
	WriteString(newCelFile, m_HeaderData.GetHeader());
	WriteString(newCelFile, m_HeaderData.GetAlg());
	WriteString(newCelFile, m_HeaderData.GetAlgorithmParameters());
	WriteUInt32_I(newCelFile, m_HeaderData.GetMargin());
	WriteUInt32_I(newCelFile, m_HeaderData.GetOutliers());
	WriteUInt32_I(newCelFile, m_HeaderData.GetMasked());
	WriteUInt32_I(newCelFile, 0);

	// Write the Mean data
	short sPixel = 0;
	for (int iCell=0; iCell < m_HeaderData.GetCells(); iCell++) {
    int   t_x=GetX(iCell);
    int   t_y=GetY(iCell);
    float t_mean=GetMean(t_x,t_y);
    float t_stdv=GetStdev(t_x,t_y);
    uint16_t t_pixel=GetPixels(t_x,t_y);

    WriteFloatLowPrecision(newCelFile,t_mean);
    WriteFloatLowPrecision(newCelFile,t_stdv);
    WriteUInt16_I(newCelFile,t_pixel);
	}

	// Write the mask data
	for(std::map<int, bool>::iterator pos = m_MaskedCells.begin(); pos != m_MaskedCells.end(); pos++)
	{
		WriteUInt16_I(newCelFile, (unsigned short) GetX(pos->first));
		WriteUInt16_I(newCelFile, (unsigned short) GetY(pos->first));
	}

	// Write the outlier data
	for(std::map<int, bool>::iterator pos = m_Outliers.begin(); pos != m_Outliers.end(); pos++)
	{
		WriteUInt16_I(newCelFile, (unsigned short) GetX(pos->first));
		WriteUInt16_I(newCelFile, (unsigned short) GetY(pos->first));
	}

	// Close the file and check the status.
	newCelFile.close();

	return !newCelFile.fail();
}

//////////////////////////////////////////////////////////////////////

bool CCELFileData::WriteTranscriptomeBCel()
{
	if (m_FullPathName.length() == 0)
	{
		SetError("No file name is set for file creation.");
		return false;
	}

	std::ofstream newCelFile;
	newCelFile.open(m_FullPathName.c_str(), std::ios::out | std::ios::binary);
	if (!newCelFile)
	{
		SetError("Unable to open the file.");
		return false;
	}
	
	m_HeaderData.SetDatHeader();

	// Write the header
	WriteFixedString(newCelFile, BCEL_HEADER_BYTES, BCEL_HEADER_LEN);
	WriteFloat_N(newCelFile, (float)m_HeaderData.GetVersion());

	WriteUInt32_N(newCelFile, STRUCT_SIZE_BCEL_HEADER_INFO);
	WriteFixedString(newCelFile, BCEL_CHUNK_HEAD, BCEL_CHUNK_NAME_SIZE);
	WriteFloat_N(newCelFile, (float)m_HeaderData.GetVersion());
	WriteUInt32_N(newCelFile, m_HeaderData.GetCols());		// columns
	WriteUInt32_N(newCelFile, m_HeaderData.GetRows());		// rows
	WriteUInt32_N(newCelFile, m_HeaderData.GetCols());		// totalx
	WriteUInt32_N(newCelFile, m_HeaderData.GetRows());		// totaly
	WriteUInt32_N(newCelFile, 0);							// offsetx
	WriteUInt32_N(newCelFile, 0);							// offsety

	GridCoordinatesType grid = m_HeaderData.GetCorners();
	WriteUInt32_N(newCelFile, grid.upperleft.x);			// GridCornerUL.x
	WriteUInt32_N(newCelFile, grid.upperleft.y);			// GridCornerUL.y
	WriteUInt32_N(newCelFile, grid.upperright.x);			// GridCornerUR.x
	WriteUInt32_N(newCelFile, grid.upperright.y);			// GridCornerUR.y
	WriteUInt32_N(newCelFile, grid.lowerleft.x);			// GridCornerLL.y
	WriteUInt32_N(newCelFile, grid.lowerleft.y);			// GridCornerLL.x
	WriteUInt32_N(newCelFile, grid.lowerright.x);			// GridCornerLR.x
	WriteUInt32_N(newCelFile, grid.lowerright.y);			// GridCornerLR.x
	WriteUInt16_N(newCelFile, 0);							// Axis-invertX
	WriteUInt16_N(newCelFile, 0);							// AxisInvertY
	WriteUInt16_N(newCelFile, 0);							// swapXY
	WriteUInt32_N(newCelFile, 0);

	int length = (int) m_HeaderData.GetDatHeader().length();		// DatHeader
	WriteUInt32_N(newCelFile, length);						
	WriteFixedString(newCelFile, BCEL_CHUNK_DTHD, BCEL_CHUNK_NAME_SIZE);
	if (length > 0)
		WriteFixedString(newCelFile, m_HeaderData.GetDatHeader(), length);
	else
	{
		SetError("Missing DatHeader.");
		Close();
		return 0;
	}
	WriteUInt32_N(newCelFile, 0);

	length = (int) m_HeaderData.GetAlg().length();					// Algorithm
	WriteUInt32_N(newCelFile, length);						
	WriteFixedString(newCelFile, BCEL_CHUNK_ALGM, BCEL_CHUNK_NAME_SIZE);
	if (length > 0)
		WriteFixedString(newCelFile, m_HeaderData.GetAlg(), length);
	else
	{
		SetError("Missing Algorithm.");
		Close();
		return 0;
	}
	WriteUInt32_N(newCelFile, 0);

	length = (int) m_HeaderData.GetAlgorithmParameters().length();	// AlgorithmParameters
	WriteUInt32_N(newCelFile, length);						
	WriteFixedString(newCelFile, BCEL_CHUNK_ALPR, BCEL_CHUNK_NAME_SIZE);
	if (length > 0)
		WriteFixedString(newCelFile, m_HeaderData.GetAlgorithmParameters(), length);
	else
	{
		SetError("Missing Algorithm Parameters.");
		Close();
		return 0;
	}
	WriteUInt32_N(newCelFile, 0);

	// Write the Mean data
	WriteUInt32_N(newCelFile,(uint32_t)(m_HeaderData.GetCells() * STRUCT_SIZE_FEATURE_DATA));
	WriteFixedString(newCelFile, BCEL_CHUNK_INTY, BCEL_CHUNK_NAME_SIZE);
	short sPixel = 0;
	for (int iCell=0; iCell < m_HeaderData.GetCells(); iCell++) {
    int   t_x=GetX(iCell);
    int   t_y=GetY(iCell);
    float t_mean=GetMean(t_x,t_y);
    float t_stdv=GetStdev(t_x,t_y);
    uint16_t t_pixel=GetPixels(t_x,t_y);

    WriteUInt16_N(newCelFile,(uint16_t)RoundNumber(t_mean));
    WriteUInt16_N(newCelFile,(uint16_t)RoundNumber(t_stdv));
    WriteUInt8(newCelFile,(uint8_t)t_pixel);
	}
	WriteUInt32_N(newCelFile, 0); // crc

	// Write the mask data
	WriteUInt32_N(newCelFile,(unsigned long)(m_HeaderData.GetMasked() * STRUCT_SIZE_XY_PAIR));
	WriteFixedString(newCelFile, BCEL_CHUNK_MASK, BCEL_CHUNK_NAME_SIZE);
	for(std::map<int, bool>::iterator pos = m_MaskedCells.begin(); pos != m_MaskedCells.end(); pos++)
	{
		WriteUInt32_N(newCelFile, (unsigned long) GetX(pos->first));
		WriteUInt32_N(newCelFile, (unsigned long) GetY(pos->first));
	}
	WriteUInt32_N(newCelFile, 0);

	// Write the outlier data
	WriteUInt32_N(newCelFile, (unsigned long) (m_HeaderData.GetOutliers() * STRUCT_SIZE_XY_PAIR));
	WriteFixedString(newCelFile, BCEL_CHUNK_OUTL, BCEL_CHUNK_NAME_SIZE);
	for(std::map<int, bool>::iterator pos = m_Outliers.begin(); pos != m_Outliers.end(); pos++)
	{
		WriteUInt32_N(newCelFile, (unsigned long) GetX(pos->first));
		WriteUInt32_N(newCelFile, (unsigned long) GetY(pos->first));
	}
	WriteUInt32_N(newCelFile, 0);

	// Write tail
	WriteUInt32_N(newCelFile, 0);
	WriteFixedString(newCelFile, BCEL_CHUNK_TAIL, BCEL_CHUNK_NAME_SIZE);
	WriteUInt32_N(newCelFile, 0);

	// Close the file and check the status.
	newCelFile.close();

	return !newCelFile.fail();
}

//////////////////////////////////////////////////////////////////////

bool CCELFileData::WriteCompactBCel()
{
	if (m_FullPathName.length() == 0)
	{
		SetError("No file name is set for file creation.");
		return false;
	}

	std::ofstream newCelFile;
	newCelFile.open(m_FullPathName.c_str(), std::ios::out | std::ios::binary);
	if (!newCelFile)
	{
		SetError("Unable to open the file.");
		return false;
	}

	m_HeaderData.SetDatHeader();

	// Write the header
	WriteFixedString(newCelFile, CCEL_HEADER_BYTES, CCEL_HEADER_LEN);
	WriteUInt32_I(newCelFile, m_HeaderData.GetVersion());

	// Write the dimensions of the array
	WriteUInt32_I(newCelFile, m_HeaderData.GetRows());
	WriteUInt32_I(newCelFile, m_HeaderData.GetCols());
	WriteUInt32_I(newCelFile, m_HeaderData.GetCells());

	// Write the other members.
	WriteString(newCelFile, m_HeaderData.GetHeader().c_str());
	WriteString(newCelFile, m_HeaderData.GetAlg().c_str());
	WriteString(newCelFile, m_HeaderData.GetAlgorithmParameters().c_str());
	WriteUInt32_I(newCelFile, m_HeaderData.GetMargin());
	WriteUInt32_I(newCelFile, 0);

	// Write the Mean data
	for (int iCell=0; iCell < m_HeaderData.GetCells(); iCell++)
	{
    int   t_x=GetX(iCell);
    int   t_y=GetY(iCell);
    float t_mean=GetMean(t_x,t_y);

    WriteUInt16_I(newCelFile,(uint16_t)RoundNumber(t_mean));
  }

	// Close the file and check the status.
	newCelFile.close();

	return !newCelFile.fail();
}

//////////////////////////////////////////////////////////////////////
