////////////////////////////////////////////////////////////////
//
// Copyright (C) 2004 Affymetrix, Inc.
//
// AFFYMETRIX, INC. LIMITED RESEARCH AND DEVELOPMENT LICENSE AGREEMENT
// 
// Affymetrix GeneChip Human Exon Array Data Analysis Software
// 		
// IMPORTANT - PLEASE READ CAREFULLY: This Limited Research and Development
// License Agreement (the Agreement) is a legally binding agreement between
// you (either an individual or an entity) and Affymetrix, Inc. for the
// version of the Affymetrix software product identified above provided in
// connection with this Agreement, which includes computer software (in
// source and/or object code form) and may include related media, printed
// materials and/or electronic documentation (the Software). This Agreement
// contains terms and conditions which are applicable to your use of the
// Software. By clicking "I Accept" below and/or accessing, downloading,
// viewing or using the Software, you are indicating that you agree to be
// bound by the following terms. If you do not agree to the terms of this
// Agreement, you are not licensed to and may not use the Software.
// 
// GRANT OF LICENSE.  Affymetrix grants you a royalty-free, non-exclusive,
// non-transferable, non-sublicensable, internal-use copyright and trade
// secret license to install and use the Software to analyze data generated
// using Affymetrix GeneChip human exon arrays, subject to all terms and
// conditions of this Agreement, the terms and conditions of the Technology
// Access Program or other agreement, if any, between you and Affymetrix (or
// one of its subsidiaries)  in connection with which the Software is
// provided, and the general terms and conditions of Affymetrix applicable to
// any product(s) to which the Software pertains. This license permits you to
// modify and make derivative works of any source code components, if any, of
// the Software and compile them into object code solely for your internal
// use consistent with the terms and conditions of this Agreement. Such
// license shall terminate within one hundred and eighty (180) calendar days
// from the day you first download or access the Software, or upon
// Affymetrix broad commercial launch, if any, of the GeneChip human exon
// array, or December 31, 2005, whichever is earlier. You agree to destroy,
// or return to Affymetrix upon Affymetrix request, all copies of the
// Software and cease using the Software upon termination of the license.
// 
// RESTRICTIONS ON SOFTWARE. You shall not: (i) modify or create derivative
// works of the Software, except as expressly permitted above for your
// internal use only; (ii) distribute, market, rent, lease or transfer to any
// third party the Software or any portion thereof; (iii) use the Software in
// any time sharing or rental arrangement; (iv) download, upload, or
// otherwise export or re-export the Software or any underlying information
// or technology (or direct product thereof) except in full compliance with
// all United States and other applicable laws and regulations; nor (v)
// remove any product identification, trademark, copyright or other notices.
// 
// INTELLECTUAL PROPERTY RIGHTS.  Subject to your limited license to use the
// Software as provided above, you agree that Affymetrix owns all copyright,
// patent, trademark, trade secret, and all other intellectual property
// rights in the Software. You hereby grant to Affymetrix a non-exclusive,
// worldwide, fully sublicensable, fully paid-up, royalty-free, irrevocable,
// perpetual license to any and all modifications or improvements you make to
// the Software code and any feedback or guidance you provide regarding the
// Software; the foregoing may be used, disclosed, disseminated and/or
// published by Affymetrix for any purpose, including incorporation into the
// Software or other software, without obligation of any kind to you.
// 	
// TRADEMARKS. No license, right, or interest in any Affymetrix trademark,
// trade name, service mark, or other designation is granted hereunder.
// 
// NO IMPLIED LICENSE. You acquire no rights or licenses, implied or
// otherwise, except those expressly and unambiguously granted herein.
// 
// NO PATENT LICENSE. Nothing contained herein shall be construed as
// conferring by implication, estoppel or otherwise any license or right
// under any patent, whether or not the exercise of any right herein granted
// necessarily employs an invention claimed by any existing or later issued
// patent.  Notwithstanding the preceding sentence, you shall have the right
// to exercise all rights expressly and unambiguously granted pursuant to
// this Agreement.
// 
// NO WARRANTY. The Software is provided AS IS and Affymetrix makes no
// warranty as to the Software. Affymetrix shall have no obligation under
// this Agreement to correct or notify you of any bugs, defects or errors in
// the Software or to otherwise support or maintain the Software. No warranty
// with respect to performance specifications of the Software is made. You
// acknowledge and agree that Affymetrix may modify or cease supporting the
// Software at any time with or without notice and the entire risk of using
// the Software is borne by you. You further acknowledge that Affymetrix has
// not promised or guaranteed to you that the Software will be commercially
// launched or made available to anyone in the future, that Affymetrix has no
// express or implied obligation to you to commercially launch or introduce
// the Software and that Affymetrix may elect not to introduce a product
// similar to or compatible with the Software. AFFYMETRIX DISCLAIMS ALL
// WARRANTIES, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE IMPLIED
// WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
// NON-INFRINGEMENT, RELATED TO THE SOFTWARE, ITS USE OR ANY INABILITY TO USE
// IT, THE RESULTS OF ITS USE AND THIS AGREEMENT.
// 
// U.S. GOVERNMENT END USERS. If you are an agency, department, or other
// entity of the United States government (Government), the use,
// duplication, reproduction, release, modification, disclosure or transfer
// of this product, or of any related documentation of any kind, including
// technical data, is restricted in accordance with Federal Acquisition
// Regulation (FAR) 12.212 for civilian agencies and Defense Federal
// Acquisition Regulation Supplement (DFARS) 227.7202 for military
// agencies. The Software is commercial computer software and commercial
// computer software documentation. The use of the Software by the Government
// is further restricted in accordance with the terms of this Agreement. The
// contractor/manufacturer is Affymetrix Inc., 3380 Central Expressway, Santa
// Clara, CA 95051.
// 
// LIMITATIONS OF LIABILITY AND DAMAGES. IN NO EVENT SHALL AFFYMETRIX BE
// LIABLE FOR ANY DAMAGES, WHETHER IN CONTRACT OR TORT (INCLUING NEGLIGENCE)
// OR UNDER ANY OTHER LEGAL THEORY, INCLUDING BUT NOT LIMITED TO DIRECT,
// CONSEQUENTIAL, SPECIAL, EXEMPLARY, INCIDENTAL AND INDIRECT DAMAGES,
// ARISING OUT OF OR IN CONNECTION WITH THIS AGREEMENT OR THE USE OR THE
// INABILITY TO USE THE SOFTWARE OR THE RESULTS OF USING THE SOFTWARE. THE
// LIABILITY OF AFFYMETRIX SHALL IN NO EVENT EXCEED FIFTY DOLLARS (U.S.
// $50.00). YOU ACKNOWLEDGE AND AGREE THAT THE RISKS OF LOSS HEREUNDER ARE
// REFLECTED IN THE PRICE OF THE SOFTWARE LICENSE, AND THAT THESE TERMS WOULD
// HAVE BEEN DIFFERENT IF THERE HAD BEEN A DIFFERENT ALLOCATION OF RISK.
// 
// NOT APPROVED FOR DIAGNOSTIC USE.  The Software is for research use only
// and not for use in diagnostic procedures.  The Software does not have the
// approval of the U.S. Food and Drug Administration (FDA), the Japanese
// Minister of Health, Labor and Welfare (MHLW) or other regulatory
// approval. No license is conveyed or implied for you to use, and you agree
// not to use, the Software or any results generated therefrom in any manner
// requiring FDA, MHLW or similar regulatory approval FDA.
// 
// GENERAL. This Agreement constitutes the entire agreement between you and
// Affymetrix with respect to the subject matter hereof and is the final,
// complete and exclusive statement of the terms of the Agreement,
// superseding all prior written and oral agreements, understandings and
// undertakings relating to the subject matter hereof. Modifications may be
// made only in writing and signed by an authorized corporate officer of
// Affymetrix. The waiver of any term or condition or any breach thereof
// shall not affect any other term or condition of this Agreement. You may
// not assign or transfer this Agreement without the prior written consent of
// Affymetrix. Without prejudice to any other rights, Affymetrix may
// terminate this Agreement if you fail to comply with the terms and
// conditions of this Agreement, the terms and conditions of the Technology
// Access Program or other agreement, if any, between you and Affymetrix (or
// one of its subsidiaries) in connection with which the Software is
// provided, or the general terms and conditions of Affymetrix applicable to
// any product(s) to which the Software pertains. In such event, you must
// destroy, or return to Affymetrix upon Affymetrix request, all copies of
// the Software and cease using the Software. This Agreement shall be
// governed by and construed according to the laws of California, without
// regard to conflict of law provisions. In the event that any provision of
// this Agreement or portion thereof is found to be illegal or unenforceable,
// the Agreement shall be construed without the unenforceable provision or
// portion thereof.
// 
// Should you have any questions concerning this license agreement, or if you
// desire to contact Affymetrix for any reason please write: Affymetrix,
// Inc., 3380 Central Expressway, Santa Clara, CA 95051, Attention: Legal
// Department; or phone 408.731.5000.
// 
// AS STATED ABOVE, USE OF THE SOFTWARE INDICATES YOUR ACCEPTANCE OF THIS
// AGREEMENT.
//
////////////////////////////////////////////////////////////////

#ifndef _CELFILEDATA_H_
#define _CELFILEDATA_H_

#include <map>
#include <string>

#if defined (WIN32)
#include <windows.h>
#define STRUCT_ALIGNMENT
#elif defined (CYGWIN)
#include <stdio.h>
#include <unistd.h>
#include <sys/mman.h>
#include <sys/types.h>
#define STRUCT_ALIGNMENT __attribute__ ((packed))
#define PAGE_SHIFT	4
#define PAGE_SIZE	(1UL << PAGE_SHIFT)
#define PAGE_MASK	(~(PAGE_SIZE-1))

#else // UNIX 
#include <stdio.h>
#include <unistd.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/user.h>
#define STRUCT_ALIGNMENT __attribute__ ((packed))
#endif


namespace affxcel
{

//////////////////////////////////////////////////////////////////////

typedef struct _CELFileEntryType
{
	float Mean STRUCT_ALIGNMENT;
	float Stdv STRUCT_ALIGNMENT;
	short Pixels STRUCT_ALIGNMENT;
} CELFileEntryType;

typedef struct _CELFileTranscriptomeEntryType
{
	unsigned short Mean STRUCT_ALIGNMENT;
	unsigned short Stdv STRUCT_ALIGNMENT;
	unsigned char Pixels STRUCT_ALIGNMENT;
} CELFileTranscriptomeEntryType;

typedef struct _CoordinatePoint
{ 
  long x; 
  long y; 
} CoordinatePoint; 

typedef struct _GridCoordinatesType
{
    CoordinatePoint upperleft;
    CoordinatePoint upperright;
    CoordinatePoint lowerright;
    CoordinatePoint lowerleft;
} GridCoordinatesType;

//////////////////////////////////////////////////////////////////////

class CCELFileHeaderData
{
private:
	int m_Magic;
	int m_Version;
	int m_nCols;
	int m_nRows;
	int m_nCells;
	std::string m_Header;
	std::string m_Alg;
	std::string m_Params;
	std::string m_ChipType;
	std::string m_DatHeader;
	int m_Margin;
	unsigned long m_nOutliers;
	unsigned long m_nMasked;
	GridCoordinatesType m_CellGrid;

	std::map<std::string, std::string> m_Parameters;
	std::map<int, std::string> m_ParameterIndices;

public:
	CCELFileHeaderData();
	~CCELFileHeaderData();

	void Clear();

	void SetMagic(int i) { m_Magic = i; }
	int GetMagic() { return m_Magic; }

	void SetVersion(int i) { m_Version = i; }
	int GetVersion() { return m_Version; }

	void SetCols(int i) { m_nCols = i; }
	int GetCols() { return m_nCols; }

	void SetRows(int i) { m_nRows = i; }
	int GetRows() { return m_nRows; }

	void SetCells(int i) { m_nCells = i; }
	int GetCells() { return m_nCells; }

	void AppendHeader(const char* p) { m_Header.append(p); }
	void SetHeader(const char* p) { m_Header = p; }
	std::string GetHeader();

	void SetAlg(const char* p) { m_Alg = p; }
	std::string GetAlg() { return m_Alg; }

	void SetParams(const char* p) { m_Params = p; }
	std::string GetParams() { return m_Params; }

	void ParseChipType();
	void SetChipType(const char* p) { m_ChipType = p; }
	std::string GetChipType() { return m_ChipType; }

	void ParseDatHeader();
	void SetDatHeader();
	void SetDatHeader(const char* p) { m_DatHeader = p; }
	std::string GetDatHeader() { return m_DatHeader; }

	void SetMargin(int i) { m_Margin = i; }
	int GetMargin() { return m_Margin; }

	void SetOutliers(unsigned long l) { m_nOutliers = l; }
	unsigned long GetOutliers() { return m_nOutliers; }
	void IncrementOutliers() { m_nOutliers++; }
	void DecrementOutliers() { m_nOutliers--; }

	void SetMasked(unsigned long l) { m_nMasked = l; }
	unsigned long GetMasked() { return m_nMasked; }
	void IncrementMasked() { m_nMasked++; }
	void DecrementMasked() { m_nMasked--; }

	void ParseAlgorithmParameters();
	void AddAlgorithmParameter(std::string& tag, std::string& value);
	void AddAlgorithmParameter(const char *tag, const char *value);
	void SetAlgorithmParameter(const char *tag, const char *value);
	std::string GetAlgorithmParameter(const char *tag);
	std::string GetAlgorithmParameters();

	void SetGrid(GridCoordinatesType grid);
	GridCoordinatesType GetCorners() { return m_CellGrid; }
	void ParseCorners();
};

//////////////////////////////////////////////////////////////////////

class CCELFileData  
{
public:
	enum
	{
		TEXT_CEL = 1,
		XDA_BCEL = 2,
		TRANSCRIPTOME_BCEL = 3,
		COMPACT_BCEL = 4
	};

	enum
	{
		CEL_ALL=1,
		CEL_DATA=2,
		CEL_OUTLIER=4,
		CEL_MASK=8
	};

public:
	// Constructor/destructor
	CCELFileData();
	~CCELFileData();

private:
	int m_FileFormat;
	std::string m_strError;
	std::string m_FullPathName;
	std::string m_LibPath;
	std::string m_DataPath;
	std::string m_FileName;
	CCELFileHeaderData m_HeaderData;
	CELFileEntryType *m_pEntries;
	CELFileTranscriptomeEntryType *m_pTransciptomeEntries;

	int m_nReadState;
	bool m_bReadMaskedCells;
	bool m_bReadOutliers;
	unsigned short *m_pMeanIntensities;
	std::map<int, bool> m_MaskedCells;
	std::map<int, bool> m_Outliers;

#ifdef WIN32
	HANDLE m_hFile;
	HANDLE m_hFileMap;
#else
	FILE* m_File;
	size_t m_MapLen;
#endif
	char  *m_lpData;
	void  *m_lpFileMap;

	bool Open(bool bReadHeaderOnly = false);
	bool ReadTextCel(bool bReadHeaderOnly = false);
	bool ReadXDABCel(bool bReadHeaderOnly = false);
	bool ReadTranscriptomeBCel(bool bReadHeaderOnly = false);
	bool ReadCompactBCel(bool bReadHeaderOnly = false);
	void DetermineFileFormat();
	unsigned short RoundNumber(float f);
	int GetX(int index);
	int GetY(int index);

public:
	// Error.
	void SetError(const char* str);
	std::string GetError() { return m_strError; }

	affxcel::CCELFileData *GetThisPtr() { return this; }

	// File name parameters.
	void SetDataPath(const char *str);
	void SetLibPath(const char *str);
	void SetFileName(const char *str);
	std::string GetFileName() const { return m_FileName; }
	std::string GetFilePath() const { return m_FullPathName; }

	// Accessors for header information.
	std::string GetHeaderKey(const char* key);
	int GetVersion() { return m_HeaderData.GetVersion(); }
	int GetCols() { return m_HeaderData.GetCols(); }
	int GetRows() { return m_HeaderData.GetRows(); }
	int GetNumCells() { return m_HeaderData.GetCells(); }
	std::string GetHeader() { return m_HeaderData.GetHeader(); }
	std::string GetAlg() { return m_HeaderData.GetAlg(); }
	std::string GetParams() { return m_HeaderData.GetAlgorithmParameters(); }
	std::string GetDatHeader() { return m_HeaderData.GetDatHeader(); }
	std::string GetChipType() { return m_HeaderData.GetChipType(); }
	int GetCellMargin() { return m_HeaderData.GetMargin(); }
	unsigned long GetNumOutliers() { return m_HeaderData.GetOutliers(); }
	unsigned long GetNumMasked() { return m_HeaderData.GetMasked(); }
	GridCoordinatesType GetCorners() { return m_HeaderData.GetCorners(); }

	// Accessors for intensity information.
	float GetMean(int x, int y);
	float GetStdev(int x, int y);
	short GetPixels(int x, int y);
	bool IsMasked(int x, int y);
	bool IsOutlier(int x, int y);

	// For reading a file.
	void Close();
	long GetFileSize();
	bool Exists();
	bool ReadHeader();
	bool Read();
	bool IsXDACompatibleFile();
	bool IsTranscriptomeBcelFile();
	bool IsCompactCelFile();
	bool ReadEx(const char *filename=0, int nState=CEL_ALL);
	int GetReadState() { return m_nReadState; };

	// For writing a new CEL file
	void Clear();
	int  GetFileFormat() { return m_FileFormat; }
	void SetFileFormat(int i);
	void SetAlgorithmName(const char *str);
	void AddAlgorithmParameter(const char *tag, const char *value);
	void SetDimensions(int rows, int cols);
	void SetChipType(const char *str);
	void SetMargin(int margin);
	void SetMean(int x, int y, float mean);
	void SetStdev(int x, int y, float stdev);
	void SetPixels(int x, int y, short pixels);
	void SetMasked(int x, int y, bool masked);
	void SetOutlier(int x, int y, bool outlier);
	bool WriteTextCel();
	bool WriteXDABCel();
	bool WriteTranscriptomeBCel();
	bool WriteCompactBCel();

	// make sure it is writeable.
	void Munmap();
	void EnsureNotMmapped();
};

//////////////////////////////////////////////////////////////////////

} // namespace

////////////////////////////////////////////////////////////////////
#endif // _CELFILEDATA_H_
