////////////////////////////////////////////////////////////////
//
// Copyright (C) 2004 Affymetrix, Inc.
//
// AFFYMETRIX, INC. LIMITED RESEARCH AND DEVELOPMENT LICENSE AGREEMENT
// 
// Affymetrix GeneChip Human Exon Array Data Analysis Software
// 		
// IMPORTANT - PLEASE READ CAREFULLY: This Limited Research and Development
// License Agreement (the Agreement) is a legally binding agreement between
// you (either an individual or an entity) and Affymetrix, Inc. for the
// version of the Affymetrix software product identified above provided in
// connection with this Agreement, which includes computer software (in
// source and/or object code form) and may include related media, printed
// materials and/or electronic documentation (the Software). This Agreement
// contains terms and conditions which are applicable to your use of the
// Software. By clicking "I Accept" below and/or accessing, downloading,
// viewing or using the Software, you are indicating that you agree to be
// bound by the following terms. If you do not agree to the terms of this
// Agreement, you are not licensed to and may not use the Software.
// 
// GRANT OF LICENSE.  Affymetrix grants you a royalty-free, non-exclusive,
// non-transferable, non-sublicensable, internal-use copyright and trade
// secret license to install and use the Software to analyze data generated
// using Affymetrix GeneChip human exon arrays, subject to all terms and
// conditions of this Agreement, the terms and conditions of the Technology
// Access Program or other agreement, if any, between you and Affymetrix (or
// one of its subsidiaries)  in connection with which the Software is
// provided, and the general terms and conditions of Affymetrix applicable to
// any product(s) to which the Software pertains. This license permits you to
// modify and make derivative works of any source code components, if any, of
// the Software and compile them into object code solely for your internal
// use consistent with the terms and conditions of this Agreement. Such
// license shall terminate within one hundred and eighty (180) calendar days
// from the day you first download or access the Software, or upon
// Affymetrix broad commercial launch, if any, of the GeneChip human exon
// array, or December 31, 2005, whichever is earlier. You agree to destroy,
// or return to Affymetrix upon Affymetrix request, all copies of the
// Software and cease using the Software upon termination of the license.
// 
// RESTRICTIONS ON SOFTWARE. You shall not: (i) modify or create derivative
// works of the Software, except as expressly permitted above for your
// internal use only; (ii) distribute, market, rent, lease or transfer to any
// third party the Software or any portion thereof; (iii) use the Software in
// any time sharing or rental arrangement; (iv) download, upload, or
// otherwise export or re-export the Software or any underlying information
// or technology (or direct product thereof) except in full compliance with
// all United States and other applicable laws and regulations; nor (v)
// remove any product identification, trademark, copyright or other notices.
// 
// INTELLECTUAL PROPERTY RIGHTS.  Subject to your limited license to use the
// Software as provided above, you agree that Affymetrix owns all copyright,
// patent, trademark, trade secret, and all other intellectual property
// rights in the Software. You hereby grant to Affymetrix a non-exclusive,
// worldwide, fully sublicensable, fully paid-up, royalty-free, irrevocable,
// perpetual license to any and all modifications or improvements you make to
// the Software code and any feedback or guidance you provide regarding the
// Software; the foregoing may be used, disclosed, disseminated and/or
// published by Affymetrix for any purpose, including incorporation into the
// Software or other software, without obligation of any kind to you.
// 	
// TRADEMARKS. No license, right, or interest in any Affymetrix trademark,
// trade name, service mark, or other designation is granted hereunder.
// 
// NO IMPLIED LICENSE. You acquire no rights or licenses, implied or
// otherwise, except those expressly and unambiguously granted herein.
// 
// NO PATENT LICENSE. Nothing contained herein shall be construed as
// conferring by implication, estoppel or otherwise any license or right
// under any patent, whether or not the exercise of any right herein granted
// necessarily employs an invention claimed by any existing or later issued
// patent.  Notwithstanding the preceding sentence, you shall have the right
// to exercise all rights expressly and unambiguously granted pursuant to
// this Agreement.
// 
// NO WARRANTY. The Software is provided AS IS and Affymetrix makes no
// warranty as to the Software. Affymetrix shall have no obligation under
// this Agreement to correct or notify you of any bugs, defects or errors in
// the Software or to otherwise support or maintain the Software. No warranty
// with respect to performance specifications of the Software is made. You
// acknowledge and agree that Affymetrix may modify or cease supporting the
// Software at any time with or without notice and the entire risk of using
// the Software is borne by you. You further acknowledge that Affymetrix has
// not promised or guaranteed to you that the Software will be commercially
// launched or made available to anyone in the future, that Affymetrix has no
// express or implied obligation to you to commercially launch or introduce
// the Software and that Affymetrix may elect not to introduce a product
// similar to or compatible with the Software. AFFYMETRIX DISCLAIMS ALL
// WARRANTIES, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE IMPLIED
// WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
// NON-INFRINGEMENT, RELATED TO THE SOFTWARE, ITS USE OR ANY INABILITY TO USE
// IT, THE RESULTS OF ITS USE AND THIS AGREEMENT.
// 
// U.S. GOVERNMENT END USERS. If you are an agency, department, or other
// entity of the United States government (Government), the use,
// duplication, reproduction, release, modification, disclosure or transfer
// of this product, or of any related documentation of any kind, including
// technical data, is restricted in accordance with Federal Acquisition
// Regulation (FAR) 12.212 for civilian agencies and Defense Federal
// Acquisition Regulation Supplement (DFARS) 227.7202 for military
// agencies. The Software is commercial computer software and commercial
// computer software documentation. The use of the Software by the Government
// is further restricted in accordance with the terms of this Agreement. The
// contractor/manufacturer is Affymetrix Inc., 3380 Central Expressway, Santa
// Clara, CA 95051.
// 
// LIMITATIONS OF LIABILITY AND DAMAGES. IN NO EVENT SHALL AFFYMETRIX BE
// LIABLE FOR ANY DAMAGES, WHETHER IN CONTRACT OR TORT (INCLUING NEGLIGENCE)
// OR UNDER ANY OTHER LEGAL THEORY, INCLUDING BUT NOT LIMITED TO DIRECT,
// CONSEQUENTIAL, SPECIAL, EXEMPLARY, INCIDENTAL AND INDIRECT DAMAGES,
// ARISING OUT OF OR IN CONNECTION WITH THIS AGREEMENT OR THE USE OR THE
// INABILITY TO USE THE SOFTWARE OR THE RESULTS OF USING THE SOFTWARE. THE
// LIABILITY OF AFFYMETRIX SHALL IN NO EVENT EXCEED FIFTY DOLLARS (U.S.
// $50.00). YOU ACKNOWLEDGE AND AGREE THAT THE RISKS OF LOSS HEREUNDER ARE
// REFLECTED IN THE PRICE OF THE SOFTWARE LICENSE, AND THAT THESE TERMS WOULD
// HAVE BEEN DIFFERENT IF THERE HAD BEEN A DIFFERENT ALLOCATION OF RISK.
// 
// NOT APPROVED FOR DIAGNOSTIC USE.  The Software is for research use only
// and not for use in diagnostic procedures.  The Software does not have the
// approval of the U.S. Food and Drug Administration (FDA), the Japanese
// Minister of Health, Labor and Welfare (MHLW) or other regulatory
// approval. No license is conveyed or implied for you to use, and you agree
// not to use, the Software or any results generated therefrom in any manner
// requiring FDA, MHLW or similar regulatory approval FDA.
// 
// GENERAL. This Agreement constitutes the entire agreement between you and
// Affymetrix with respect to the subject matter hereof and is the final,
// complete and exclusive statement of the terms of the Agreement,
// superseding all prior written and oral agreements, understandings and
// undertakings relating to the subject matter hereof. Modifications may be
// made only in writing and signed by an authorized corporate officer of
// Affymetrix. The waiver of any term or condition or any breach thereof
// shall not affect any other term or condition of this Agreement. You may
// not assign or transfer this Agreement without the prior written consent of
// Affymetrix. Without prejudice to any other rights, Affymetrix may
// terminate this Agreement if you fail to comply with the terms and
// conditions of this Agreement, the terms and conditions of the Technology
// Access Program or other agreement, if any, between you and Affymetrix (or
// one of its subsidiaries) in connection with which the Software is
// provided, or the general terms and conditions of Affymetrix applicable to
// any product(s) to which the Software pertains. In such event, you must
// destroy, or return to Affymetrix upon Affymetrix request, all copies of
// the Software and cease using the Software. This Agreement shall be
// governed by and construed according to the laws of California, without
// regard to conflict of law provisions. In the event that any provision of
// this Agreement or portion thereof is found to be illegal or unenforceable,
// the Agreement shall be construed without the unenforceable provision or
// portion thereof.
// 
// Should you have any questions concerning this license agreement, or if you
// desire to contact Affymetrix for any reason please write: Affymetrix,
// Inc., 3380 Central Expressway, Santa Clara, CA 95051, Attention: Legal
// Department; or phone 408.731.5000.
// 
// AS STATED ABOVE, USE OF THE SOFTWARE INDICATES YOUR ACCEPTANCE OF THIS
// AGREEMENT.
//
////////////////////////////////////////////////////////////////

#ifndef AFFX_FILEIO_H
#define AFFX_FILEIO_H 1

//
#include <fstream>
#include <string>
//
#include "affy-base-types.h"

// Contrary to normal, Affy uses little-endian byte order
// in their files.  This requires defining a the complements 
// to the normal functions "hton".

// FIXME: need to test for sparc, not sun
#ifdef sun
#define IS_BIG_ENDIAN 1
#endif

#ifdef IS_BIG_ENDIAN
// the shuffle
#define affy_swap32(x) \
     ((((x) & 0xff000000) >> 24) | (((x) & 0x00ff0000) >>  8) | \
      (((x) & 0x0000ff00) <<  8) | (((x) & 0x000000ff) << 24))
#define affy_swap16(x) \
     ((((x) >> 8) & 0xff) | (((x) & 0xff) << 8))
// the four operators
#define htoil(x) affy_swap32(x)
#define htois(x) affy_swap16(x)
#define itohl(x) affy_swap32(x)
#define itohs(x) affy_swap16(x)
#else
// Identity operators
#define htoil(x) (x)
#define htois(x) (x)
#define itohl(x) (x)
#define itohs(x) (x)
#endif

#ifndef IFSTREAM
#define IFSTREAM std::istream
#endif 

//
void ReadUInt32_I(IFSTREAM& in,uint32_t& val);
void ReadInt32_I(IFSTREAM&  in,int32_t& val);
void ReadFloat_I(IFSTREAM& in,float& val);
void ReadUInt16_I(IFSTREAM& in,uint16_t& val);
void ReadInt16_I(IFSTREAM& in,int16_t& val);
//
void ReadUInt8(IFSTREAM& in,uint8_t& val);
void ReadInt8(IFSTREAM& in,int8_t& val);
// The network byte order.
void ReadUInt32_N(IFSTREAM& in,uint32_t& val);
void ReadInt32_N(IFSTREAM&  in,int32_t& val);
void ReadFloat_N(IFSTREAM& in,float& val);
void ReadUInt16_N(IFSTREAM& in,uint16_t& val);
void ReadInt16_N(IFSTREAM& in,int16_t& val);
//
void ReadFixedCString(IFSTREAM& in, char* str, uint32_t len);
void ReadCString(IFSTREAM& in, char*& str);
void ReadFixedString(IFSTREAM& in, std::string& str, uint32_t len);
void ReadString(IFSTREAM& in, std::string& str);
//
void ReadNextLine(IFSTREAM& in,char* line,int len);

//==============================

// Memory Map Accessors
// this is where we handle byteorder and alignment.

uint32_t MmGetUInt32_I(uint32_t* ptr);
int32_t  MmGetInt32_I(int32_t* ptr);
uint16_t MmGetUInt16_I(uint16_t* ptr);
int16_t  MmGetInt16_I(int16_t* ptr);
//
uint8_t  MmGetUInt8(uint8_t*   ptr);
int8_t   MmGetInt8(int8_t*   ptr);
//
uint32_t MmGetUInt32_N(uint32_t* ptr);
uint16_t MmGetUInt16_N(uint16_t* ptr);
float    MmGetFloat_N(float*     ptr);
int32_t  MmGetInt32_N(int32_t* ptr);
int16_t  MmGetInt16_N(int16_t* ptr);
//
float    MmGetFloat_I(float*     ptr);
float    MmGetFloat_N(float*     ptr);

//
void MmSetUInt32_I(uint32_t* ptr,uint32_t val);
void MmSetUInt16_I(uint16_t* ptr,uint16_t val);
void MmSetUInt8(uint8_t*   ptr,uint8_t  val);
void MmSetFloat_I(float*     ptr,float    val);
//
void MmSetUInt32_N(uint32_t* ptr,uint32_t val);
void MmSetUInt16_N(uint16_t* ptr,uint16_t val);
void MmSetFloat_N(float*     ptr,float    val);


//==============================
//
// COMPATABLITY DEFS - calls to these should be replaced.
//

#define LONG_T      int32_t
#define ULONG_T     uint32_t
#define FLOAT_SIZE  sizeof(float)
#define INT_SIZE    sizeof(int32_t)
#define LONG_SIZE   sizeof(int32_t)
#define ULONG_SIZE  sizeof(uint32_t)
#define SHORT_SIZE  sizeof(int16_t)
#define USHORT_SIZE sizeof(uint16_t)
#define CHAR_SIZE   sizeof(int8_t)
#define UCHAR_SIZE  sizeof(uint8_t)

//
#define READ_INT(s,v)     ReadInt32_I(s,v)
#define READ_LONG(s,v)    ReadInt32_I(s,v)
#define READ_ULONG(s,v)   ReadUInt32_I(s,v)
#define READ_SHORT(s,v)   ReadInt16_I(s,v)
#define READ_USHORT(s,v)  ReadUInt16_I(s,v)
#define READ_CHAR(s,v)    ReadInt8(s,v)
#define READ_UCHAR(s,v)   ReadUInt8(s,v)
#define READ_BOOL(s,v)    ReadInt8(s,v)
#define READ_FLOAT(s,v)   ReadFloat(s,v)

//
#define READ_STRING(s,v)         ReadCString(s,v)
#define READ_FIXED_STRING(s,v,l) ReadFixedCString(s,v,l)
#define GET_NEXT_LINE(s,v)       ReadNextLine(s,v)

// And what are these for?
#define WINDOWS_BINARY 2
#define UNIX_BINARY    3

// void ReadFixedString(IFSTREAM& instr, std::string& s, int len);
// void ReadFloat(IFSTREAM& instr, float& fval);
// void ReadString(IFSTREAM& instr, std::string& s);

#define ReadULong(s,v)    ReadUInt32_I(s,v)
#define ReadUShort(s,v)   ReadUInt16_I(s,v)
#define ReadUChar(s,v)    ReadUInt8(s,v)
#define ReadFloat(s,v)    ReadFloat_I(s,v)

//
#define ReadFixedUChar(s,v,l)   ReadFixedCString(s,v,l)
#define ReadULongLenString(s,v) ReadString(s,v)


//==============================
// MMap funcs
//
#define GetLong(p,f)            ((int32_t)MmGetUInt32_I((uint32_t*)p))
#define GetULong(p,f)           (MmGetUInt32_I(p))
#define GetShort(p,f)           ((int16_t)MmGetUInt16_I((uint16_t*)p))
#define GetUShort(p,f)          (MmGetUInt16_I(p,f))
#define GetFloat(p,f)           (MmGetFloat(p))

#endif // AFFX_FILEIO_H
