////////////////////////////////////////////////////////////////
//
// Copyright (C) 2004 Affymetrix, Inc.
//
// AFFYMETRIX, INC. LIMITED RESEARCH AND DEVELOPMENT LICENSE AGREEMENT
// 
// Affymetrix GeneChip Human Exon Array Data Analysis Software
// 		
// IMPORTANT - PLEASE READ CAREFULLY: This Limited Research and Development
// License Agreement (the Agreement) is a legally binding agreement between
// you (either an individual or an entity) and Affymetrix, Inc. for the
// version of the Affymetrix software product identified above provided in
// connection with this Agreement, which includes computer software (in
// source and/or object code form) and may include related media, printed
// materials and/or electronic documentation (the Software). This Agreement
// contains terms and conditions which are applicable to your use of the
// Software. By clicking "I Accept" below and/or accessing, downloading,
// viewing or using the Software, you are indicating that you agree to be
// bound by the following terms. If you do not agree to the terms of this
// Agreement, you are not licensed to and may not use the Software.
// 
// GRANT OF LICENSE.  Affymetrix grants you a royalty-free, non-exclusive,
// non-transferable, non-sublicensable, internal-use copyright and trade
// secret license to install and use the Software to analyze data generated
// using Affymetrix GeneChip human exon arrays, subject to all terms and
// conditions of this Agreement, the terms and conditions of the Technology
// Access Program or other agreement, if any, between you and Affymetrix (or
// one of its subsidiaries)  in connection with which the Software is
// provided, and the general terms and conditions of Affymetrix applicable to
// any product(s) to which the Software pertains. This license permits you to
// modify and make derivative works of any source code components, if any, of
// the Software and compile them into object code solely for your internal
// use consistent with the terms and conditions of this Agreement. Such
// license shall terminate within one hundred and eighty (180) calendar days
// from the day you first download or access the Software, or upon
// Affymetrix broad commercial launch, if any, of the GeneChip human exon
// array, or December 31, 2005, whichever is earlier. You agree to destroy,
// or return to Affymetrix upon Affymetrix request, all copies of the
// Software and cease using the Software upon termination of the license.
// 
// RESTRICTIONS ON SOFTWARE. You shall not: (i) modify or create derivative
// works of the Software, except as expressly permitted above for your
// internal use only; (ii) distribute, market, rent, lease or transfer to any
// third party the Software or any portion thereof; (iii) use the Software in
// any time sharing or rental arrangement; (iv) download, upload, or
// otherwise export or re-export the Software or any underlying information
// or technology (or direct product thereof) except in full compliance with
// all United States and other applicable laws and regulations; nor (v)
// remove any product identification, trademark, copyright or other notices.
// 
// INTELLECTUAL PROPERTY RIGHTS.  Subject to your limited license to use the
// Software as provided above, you agree that Affymetrix owns all copyright,
// patent, trademark, trade secret, and all other intellectual property
// rights in the Software. You hereby grant to Affymetrix a non-exclusive,
// worldwide, fully sublicensable, fully paid-up, royalty-free, irrevocable,
// perpetual license to any and all modifications or improvements you make to
// the Software code and any feedback or guidance you provide regarding the
// Software; the foregoing may be used, disclosed, disseminated and/or
// published by Affymetrix for any purpose, including incorporation into the
// Software or other software, without obligation of any kind to you.
// 	
// TRADEMARKS. No license, right, or interest in any Affymetrix trademark,
// trade name, service mark, or other designation is granted hereunder.
// 
// NO IMPLIED LICENSE. You acquire no rights or licenses, implied or
// otherwise, except those expressly and unambiguously granted herein.
// 
// NO PATENT LICENSE. Nothing contained herein shall be construed as
// conferring by implication, estoppel or otherwise any license or right
// under any patent, whether or not the exercise of any right herein granted
// necessarily employs an invention claimed by any existing or later issued
// patent.  Notwithstanding the preceding sentence, you shall have the right
// to exercise all rights expressly and unambiguously granted pursuant to
// this Agreement.
// 
// NO WARRANTY. The Software is provided AS IS and Affymetrix makes no
// warranty as to the Software. Affymetrix shall have no obligation under
// this Agreement to correct or notify you of any bugs, defects or errors in
// the Software or to otherwise support or maintain the Software. No warranty
// with respect to performance specifications of the Software is made. You
// acknowledge and agree that Affymetrix may modify or cease supporting the
// Software at any time with or without notice and the entire risk of using
// the Software is borne by you. You further acknowledge that Affymetrix has
// not promised or guaranteed to you that the Software will be commercially
// launched or made available to anyone in the future, that Affymetrix has no
// express or implied obligation to you to commercially launch or introduce
// the Software and that Affymetrix may elect not to introduce a product
// similar to or compatible with the Software. AFFYMETRIX DISCLAIMS ALL
// WARRANTIES, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE IMPLIED
// WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
// NON-INFRINGEMENT, RELATED TO THE SOFTWARE, ITS USE OR ANY INABILITY TO USE
// IT, THE RESULTS OF ITS USE AND THIS AGREEMENT.
// 
// U.S. GOVERNMENT END USERS. If you are an agency, department, or other
// entity of the United States government (Government), the use,
// duplication, reproduction, release, modification, disclosure or transfer
// of this product, or of any related documentation of any kind, including
// technical data, is restricted in accordance with Federal Acquisition
// Regulation (FAR) 12.212 for civilian agencies and Defense Federal
// Acquisition Regulation Supplement (DFARS) 227.7202 for military
// agencies. The Software is commercial computer software and commercial
// computer software documentation. The use of the Software by the Government
// is further restricted in accordance with the terms of this Agreement. The
// contractor/manufacturer is Affymetrix Inc., 3380 Central Expressway, Santa
// Clara, CA 95051.
// 
// LIMITATIONS OF LIABILITY AND DAMAGES. IN NO EVENT SHALL AFFYMETRIX BE
// LIABLE FOR ANY DAMAGES, WHETHER IN CONTRACT OR TORT (INCLUING NEGLIGENCE)
// OR UNDER ANY OTHER LEGAL THEORY, INCLUDING BUT NOT LIMITED TO DIRECT,
// CONSEQUENTIAL, SPECIAL, EXEMPLARY, INCIDENTAL AND INDIRECT DAMAGES,
// ARISING OUT OF OR IN CONNECTION WITH THIS AGREEMENT OR THE USE OR THE
// INABILITY TO USE THE SOFTWARE OR THE RESULTS OF USING THE SOFTWARE. THE
// LIABILITY OF AFFYMETRIX SHALL IN NO EVENT EXCEED FIFTY DOLLARS (U.S.
// $50.00). YOU ACKNOWLEDGE AND AGREE THAT THE RISKS OF LOSS HEREUNDER ARE
// REFLECTED IN THE PRICE OF THE SOFTWARE LICENSE, AND THAT THESE TERMS WOULD
// HAVE BEEN DIFFERENT IF THERE HAD BEEN A DIFFERENT ALLOCATION OF RISK.
// 
// NOT APPROVED FOR DIAGNOSTIC USE.  The Software is for research use only
// and not for use in diagnostic procedures.  The Software does not have the
// approval of the U.S. Food and Drug Administration (FDA), the Japanese
// Minister of Health, Labor and Welfare (MHLW) or other regulatory
// approval. No license is conveyed or implied for you to use, and you agree
// not to use, the Software or any results generated therefrom in any manner
// requiring FDA, MHLW or similar regulatory approval FDA.
// 
// GENERAL. This Agreement constitutes the entire agreement between you and
// Affymetrix with respect to the subject matter hereof and is the final,
// complete and exclusive statement of the terms of the Agreement,
// superseding all prior written and oral agreements, understandings and
// undertakings relating to the subject matter hereof. Modifications may be
// made only in writing and signed by an authorized corporate officer of
// Affymetrix. The waiver of any term or condition or any breach thereof
// shall not affect any other term or condition of this Agreement. You may
// not assign or transfer this Agreement without the prior written consent of
// Affymetrix. Without prejudice to any other rights, Affymetrix may
// terminate this Agreement if you fail to comply with the terms and
// conditions of this Agreement, the terms and conditions of the Technology
// Access Program or other agreement, if any, between you and Affymetrix (or
// one of its subsidiaries) in connection with which the Software is
// provided, or the general terms and conditions of Affymetrix applicable to
// any product(s) to which the Software pertains. In such event, you must
// destroy, or return to Affymetrix upon Affymetrix request, all copies of
// the Software and cease using the Software. This Agreement shall be
// governed by and construed according to the laws of California, without
// regard to conflict of law provisions. In the event that any provision of
// this Agreement or portion thereof is found to be illegal or unenforceable,
// the Agreement shall be construed without the unenforceable provision or
// portion thereof.
// 
// Should you have any questions concerning this license agreement, or if you
// desire to contact Affymetrix for any reason please write: Affymetrix,
// Inc., 3380 Central Expressway, Santa Clara, CA 95051, Attention: Legal
// Department; or phone 408.731.5000.
// 
// AS STATED ABOVE, USE OF THE SOFTWARE INDICATES YOUR ACCEPTANCE OF THIS
// AGREEMENT.
//
////////////////////////////////////////////////////////////////

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "FileWriter.h"

//////////////////////////////////////////////////////////////////////

#ifdef WIN32
#include <winsock2.h>
#else
#include <inttypes.h>
#include <netinet/in.h>
#include <sys/types.h>
#endif

using namespace std;

//
void 
WriteUInt32_I(OFSTREAM& out,uint32_t val)
{
  val = htoil(val);
 	out.write((char *)&val,sizeof(val));
}
void
WriteInt32_I(OFSTREAM& out,int32_t val)
{
  WriteUInt32_I(out,(uint32_t)val);
}
//
void 
WriteUInt32_N(OFSTREAM& out,uint32_t val)
{
  val = htonl(val);
 	out.write((char *)&val,sizeof(val));
}
void
WriteInt32_N(OFSTREAM& out,int32_t val)
{
  WriteUInt32_N(out,(uint32_t)val);
}

//
void
WriteUInt16_I(OFSTREAM& out,uint16_t val)
{
  val = htois(val);
 	out.write((char *)&val,sizeof(val));
}
void
WriteInt16_I(OFSTREAM& out,int16_t val)
{
  WriteUInt16_I(out,(uint16_t)val);
}
void
WriteUInt16_N(OFSTREAM& out,uint16_t val)
{
  val = htons(val);
 	out.write((char *)&val,sizeof(val));
}
void
WriteInt16_N(OFSTREAM& out,int16_t val)
{
  WriteUInt16_N(out,(uint16_t)val);
}

//
void 
WriteUInt8(OFSTREAM& out,uint8_t val)
{
 	out.write((char *)&val,sizeof(val));
}
void 
WriteInt8(OFSTREAM& out,int8_t val)
{
  WriteUInt8(out,(uint8_t)val);
}

//
void 
WriteFloat_I(OFSTREAM& out,float fval)
{
	WriteUInt32_I(out,*(uint32_t *)(&fval));
}
void 
WriteFloat_N(OFSTREAM& out,float fval)
{
	WriteUInt32_N(out,*(uint32_t *)(&fval));
}


// Yes, boys and girls we are throwing away the low order bits.
// Story has it that customers wanted repeatablity with old
// text files with only had a single digit after the decimal point.

void
WriteFloatLowPrecision(OFSTREAM& out,float fval)
{
  fval=(float)((floor((fval+0.05)*10))/10.0);
	WriteFloat_I(out,fval);
}

//====================

// normal C strings.
void 
WriteFixedCString(OFSTREAM& out,const char* str,size_t outlen)
{
  uint32_t slen=(uint32_t)strlen(str);
  // printf("writefixedcstring: len('%s')=%d outlen=%d\n",str,slen,outlen); //jhg
  out.write(str,((outlen<slen)?(uint32_t)outlen:slen));
  // pad with nulls
  for (uint32_t i=slen;i<outlen;i++) {
    // printf("XXX-"); //jhg
    out.put(0);
  }
}
void 
WriteCString(OFSTREAM& out,const char* str)
{
  uint32_t slen=(uint32_t)strlen(str);
 	WriteUInt32_I(out,slen);
  WriteFixedCString(out,str,slen);
}

// C++ strings
void 
WriteFixedString(OFSTREAM& out,const string& str,size_t outlen)
{
  WriteFixedCString(out,str.c_str(),outlen);
}
void
WriteString(OFSTREAM& out,const string& str)
{
  WriteCString(out,str.c_str());
}
